/*
 * Copyright 2012 Miguel Mendoza - miguel@micovery.com
 *
 * Insane Punisher is free software: you can redistribute it and/or modify it under the terms of the 
 * GNU General Public License as published by the Free Software Foundation, either version 3 of the License, 
 * or (at your option) any later version. Insane Punisher is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the 
 * GNU General Public License along with Insane Punisher. If not, see http://www.gnu.org/licenses/.
 * 
 */


using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections.Generic;
using System.Collections;
using System.Net;
using System.Net.Mail;
using System.Web;
using System.Data;

using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;
using System.CodeDom.Compiler;
using System.Reflection;
using Microsoft.CSharp;
using System.CodeDom;


using PRoCon.Core;
using PRoCon.Core.Plugin;
using PRoCon.Core.Plugin.Commands;
using PRoCon.Core.Players;
using PRoCon.Core.Players.Items;
using PRoCon.Core.Battlemap;
using PRoCon.Core.Maps;



namespace PRoConEvents
{

    public class InsanePunisher : PRoConPluginAPI, IPRoConPluginInterface
    {

        /*
        ######################################################################################
        ##   Team kill warnings 
        ######################################################################################
        ####
        #### This is list of warnings that are sent to team killer. 
        #### For each team kill, one warning is sent to the player who violated the rules. 
        ####
        #### You can have as many warnings as you want. 
        #### They are sent in the same order as listed.
        ####
        #### The following replacements can be used:
        ####
        #### %killer% - the name of the soldier that team killed
        #### %victim% - the name of the soldier that was team killed
        #### %count%  - the number of team kills by the killer
        ####
        */


        List<string> default_warn_team_kill = new List<string>(new string[]
        {
            "%killer%, no team killing please",
            "%killer%, you have two team kills!",
            "%killer%, you have %count% team kills!"
        });


        /*
         ######################################################################################
         ##   Slaps 
         ######################################################################################
         ####
         #### This is a list of slaps that users can give each other with the '!slap' command.
         #### You can have as many types of slaps as you can come up with. Slaps are chosen at random.
         #### Slaps are publicly annouced.
         ####
         #### The variable 'slap_player' controls whether insulting is allowed or not.
         ####
        */

        List<string> default_slaps = new List<string>(new string[]
        {
             "%sender% slapped %receiver% with a big smelly tuna fish!",
             "%sender% bitch slapped %receiver%!",
             "%sender% slapped %receiver% to the back of the head!",
             "%sender% ninja slapped %receiver%!",
             "%sender% pimp slapped %receiver%!",
             "%sender% cheek slapped %receiver%!",
             "%sender% butt slapped %receiver%!",
             "%sender% slapped %receiver% in the face with a rotten old fish!",
             "%sender% slapped %receiver% with a large trout!",
             "%sender% slapped %receiver% with the remains of a big Smelly Trout!",
             "%sender% slapped %receiver% with a big bucket of fish!",
             "%sender% slapped %receiver% with a silk glove!",
             "%sender% slapped %receiver% with a wet noddle!",
             "%sender% slapped %receiver% to wake him up!",
             "%sender% slapped %receiver% with a fluffy pillow!",
             "%sender% slapped %receiver% with a rubber chicken!",
             "%sender% slapped %receiver%'s face so hard, %receiver% has to walk backwards from now on to see where he/she is going!",
             "%sender% slapped some sense into %receiver% with a red brick!",
             "%sender% slapped %receiver% with a fire hose!",
             "%sender% slapped %receiver% with a huge law suit!",
             "%sender% slapped %receiver% with a great big, wet, 100% rubber duck!",
             "%sender% slapped %receiver% with a gallon of hot fudge covered ice-cream topped with peanut-butter M&M's!",
             "%sender% slapped %receiver% with a large squid ... I hope you like seafood!",
             "%sender% slapped %receiver% with a soft feather pillow full of drake feathers!",
             "%sender% slapped %receiver% on the butt really hard!",
             "%sender% slapped %receiver% in the nuts!",
             "%sender% slapped %receiver% ... with a large black dildo!"
        });

        /*
        ######################################################################################
        ##   Insults 
        ######################################################################################
        ####
        #### This is a list of insults that users can send to each other with the '!insult' command
        #### You can have as many insults as you want. Insults are chosen at random.  Only the sender 
        #### and the reciever of the insult see the message in the chat box.
        ####
        #### The variable 'insult_player' controls whether insulting is allowed or not.
        ####
        */

        List<string> default_insults = new List<string>(new string[]
        {
            "%name%, don't go away mad. Just go away.",
            "%name%, I'd like to help you out ... which way did you come in.",
            "%name%, you are so down to earth but not quite far down enough.",
            "%name%, you are living proof that manure can grow legs and walk.",
            "%name%, you are no longer beneath my contempt.",
            "%name%, tell me where your family tree is I'd like to cut it down!",
            "%name%, when I look into your eyes, I see the back of your head.",
            "%name%, I reprimanded my son for mimicking you. I told him not to act like a fool.",
            "%name%, make a mental note ... oh, I see you're out of paper!",
            "%name%, I hear you changed your mind! And got back two cents.",
            "%name%, I heard you got a brain transplant and the brain rejected you!",
            "%name%, I heard the doctors examined your head and found nothing.",
            "%name%, I know you are nobody's fool. You're everybody's.",
            "%name%, you are so ugly, the tide runs away from you.",
            "%name% is so ugly he went to the zoo and they kept him.",
            "%name% is so so ugly, he went to a freak show and got a permanent job.",
            "%name% fell out of the ugly tree, you hit every branch on the way down.",
            "%name%, in the morning I will be sober, but you will still be ugly!",
            "%name%, Is it Halloween? Oh ... that's your real face.",
            "%name%, don't turn the other cheek. It's just as ugly.",
            "%name%, your'so ugly the only place you're ever invited is outside."
        });

        /*
        ######################################################################################
        ##   Mocks 
        ######################################################################################
        ####
        #### This is a list of mockeries that are sent to soldiers when they suicide. 
        #### The suicide mockery message is chosen at random. 
        #### 
        #### The variable 'mock_suicide' controls whether soldiers are mocked on suicide or not.
        ####
        */

        List<string> default_mocks = new List<string>(new string[]
        {
            "%name% suicided, epic fail!",
            "%name% suicided, are you even trying soldier!",
            "%name% suicided, press Alt-F4 for auto-revive! "
        });



        public Dictionary<string, bool> booleanVariables;
        public Dictionary<string, int> integerVariables;
        public Dictionary<string, float> floatVariables;
        public Dictionary<string, string> stringListVariables;
        public Dictionary<string, string> stringVariables;
        public Dictionary<string, string> enumVariables;
        public Dictionary<string, Type> enumVariablesType;
        public Dictionary<string, List<string>> variableGroups;
        public List<string> variableGroupsOrder;
        public List<string> hiddenVariables;


        private Dictionary<string, PlayerProfile> players;


        public enum ViolationType { team_kill, invalid };
        public enum ActionType { kick, punish, forgive, invalid, ban };
        public enum AutoAction { None, Kick, Ban };

        /* - limbo: when the plugin is enabled for the first, time we do not know the state of the players if they are dead or alive
         * - left/kicked: after a player leaves, there might still be other objects referencing the object of the player that just left,
         *                and it's necessary to know that the player is not in-game anymore.
         * - violated: this is a special case of being dead. when you are violated, the victim died because the killer violated the rules.
         *             the purpose of this state is to avoid counting the same violation multiple times against the killer. (specially for overlapping base zones)
         */

        public enum PlayerState { dead, alive, left, kicked, limbo, violated, banned };
        public enum BanDuration { Temporary, Permanent, Round };
        public enum BanType { GUID, IPAddress, Name };
        public enum BooleanTrueFalse { True, False };



        private class PlayerPunishment
        {
            public PlayerProfile punisher;
            public PlayerProfile punished;
            public List<ViolationType> violations;

            public PlayerPunishment(PlayerProfile pr, PlayerProfile pd, List<ViolationType> v)
            {
                punisher = pr;
                punished = pd;
                violations = v;
            }
        }


        private class PlayerMessage
        {
            public string text;
            public int time;


            public PlayerMessage(string tx)
            {
                text = tx;
            }
        }

        private class PlayerProfile
        {
            private InsanePunisher plugin;
            public string name;
            public PlayerState state;
            public CPlayerInfo info;
            public CPunkbusterInfo pbinfo;
            public Queue<PlayerMessage> qmsg;          //queued messages 
            public Queue<PlayerPunishment> qpun;       //queued punishments
            public Dictionary<PlayerProfile, int> tks; //TeamKills
            public Dictionary<PlayerProfile, int> vtk; //Vote kicks
            public Dictionary<PlayerProfile, int> vtb; //Vote bans

            public Dictionary<PlayerProfile, int> atks; //accumulated TeamKills

            public PlayerProfile killer;

            public PlayerProfile(InsanePunisher plg, CPlayerInfo inf)
            {
                info = inf;
                pbinfo = new CPunkbusterInfo();
                name = info.SoldierName;
                plugin = plg;
                resetStats();
            }

            public PlayerProfile(InsanePunisher plg, CPunkbusterInfo inf)
            {
                info = new CPlayerInfo();
                pbinfo = inf;
                name = pbinfo.SoldierName;
                plugin = plg;
                resetStats();
            }

            public PlayerProfile(InsanePunisher plg, string nm)
            {
                name = nm;
                info = new CPlayerInfo();
                pbinfo = new CPunkbusterInfo();
                plugin = plg;
                resetStats();
            }


            public bool isUS()
            {
                return info.TeamID == 1;
            }

            public bool isRU()
            {
                return info.TeamID == 2;
            }

            public void resetStats()
            {
                //Pending (team kill)
                tks = new Dictionary<PlayerProfile, int>();


                //Accumulated (team kill)
                atks = new Dictionary<PlayerProfile, int>();

                //votes
                vtk = new Dictionary<PlayerProfile, int>();

                //queued punishments
                qpun = new Queue<PlayerPunishment>();

                //queued messages
                qmsg = new Queue<PlayerMessage>();

                //bans
                vtb = new Dictionary<PlayerProfile, int>();

                //other
                killer = null;
                state = PlayerState.limbo;
            }

            public void dequeueMessages()
            {

                while (this.qmsg.Count > 0)
                {
                    PlayerMessage msg = this.qmsg.Dequeue();
                    this.plugin.SendPlayerMessage(name, msg.text);
                }
            }

            public void enqueueMessage(PlayerMessage msg)
            {
                this.qmsg.Enqueue(msg);
            }

            public override string ToString()
            {
                return ToString(false);
            }

            public string ToString(bool long_format)
            {
                String format = String.Empty;

                if (long_format)
                    format = "{0,-15} team_kills: {1,-3}, vote_kicks: {2,-3}, vote_bans: {3,-3} state: {4,-5}";
                else
                    format = "{0,-15} tk: {1,-3}, vk: {2,-3}, vb: {3,-3} s: {4,-5}";

                return String.Format(format, name, getTeamKills(), getVotesReceived(), getBanVotesReceived(), state.ToString());
            }

            public void addTk(PlayerProfile victim)
            {
                if (!tks.ContainsKey(victim))
                    tks.Add(victim, 0);

                if (victim.isViolated())
                    return;
                else
                    victim.violatedBy(this);

                tks[victim]++;

                if (this.state.Equals(PlayerState.limbo))
                    this.state = PlayerState.alive;

                plugin.showActionCommands(victim, this, ViolationType.team_kill);
            }

            public void violatedBy(PlayerProfile killer)
            {
                this.state = PlayerState.violated;
                this.killedBy(killer);
            }

            public void killedBy(PlayerProfile k)
            {
                this.killer = k;
            }

            public PlayerProfile getLastKiller()
            {
                return this.killer;
            }

            public bool forgivenBy(PlayerProfile victim)
            {
                int violations = getPendingViolations(victim);

                if (violations == 0)
                {
                    //user cannot be forgiven unless a violation exists against the victim
                    plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + "you cannot forgive " + this.name + ", unless for " + plugin.getActionableViolationsStr(ActionType.forgive));
                    return false;
                }

                plugin.SendGlobalMessage(victim.name + " forgave " + this.name + " for " + getPendingViolationsStr(victim));
                accumulateV(victim);
                return true;
            }

            public bool punishedBy(PlayerProfile victim, Boolean auto)
            {
                if (auto && plugin.isImmune(this))
                    return false;

                if (!auto && plugin.isImmune(this) && !plugin.isImmune(victim))
                {
                    plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + "you cannot punish " + this.name + " because he has immunity");
                    return false;
                }

                int violations = getPendingViolations(victim);
                if (violations == 0)
                {
                    //user cannot be punished unless a violation exists against the victim
                    if (!auto)
                        plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + "you cannot punish " + this.name + ", unless for " + plugin.getActionableViolationsStr(ActionType.punish));
                    return false;
                }

                if (this.isAlive())
                {
                    List<ViolationType> auto_list = new List<ViolationType>();
                    List<ViolationType> full_list = getPendingViolationsList(victim);

                    /* remove the auto violations */
                    full_list.RemoveAll(delegate(ViolationType item)
                    {
                        if (plugin.getBooleanVarValue("auto_punish_" + item.ToString()))
                        {
                            auto_list.Add(item);
                            return true;
                        }
                        return false;
                    });

                    if (auto && auto_list.Count > 0)
                    {
                        string violations_str = plugin.list2string(violationsList2StringList(auto_list), "and");
                        plugin.SendGlobalMessage(this.name + " was auto punished for " + violations_str + " " + victim.name);
                    }
                    else if (!auto && full_list.Count > 0)
                    {
                        string violations_str = plugin.list2string(violationsList2StringList(auto_list), "and");
                        plugin.SendGlobalMessage(victim.name + " punished " + this.name + " for " + getPendingViolationsStr(victim));
                    }

                    plugin.KillPlayer(this.name);
                }
                else if (this.isDead())
                {
                    if (auto)
                        plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + this.name + " " + plugin.playerstate2stringED(this.state) + ", auto punishment was queued.");
                    else
                        plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + this.name + " " + plugin.playerstate2stringED(this.state) + ", your punishment was queued.");

                    queuePunishmentFrom(victim);
                }
                else
                {
                    if (auto)
                        plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + this.name + " " + plugin.playerstate2stringED(this.state) + ", and cannot be auto punished.");
                    else
                        plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + this.name + " " + plugin.playerstate2stringED(this.state) + ", and cannot be punished.");
                    return false;
                }

                accumulateV(victim);
                return true;
            }

            private void queuePunishmentFrom(PlayerProfile victim)
            {
                qpun.Enqueue(new PlayerPunishment(victim, this, getPendingViolationsList(victim)));
            }

            public void dequeuePunishment()
            {

                if (qpun.Count == 0)
                    return;

                PlayerPunishment pp = qpun.Dequeue();

                List<string> list = new List<string>();
                foreach (ViolationType violation in pp.violations)
                    list.Add(plugin.violation2stringING(violation));

                string incidents = plugin.list2string(list);

                plugin.SendGlobalMessage(pp.punisher.name + " punished " + pp.punished.name + " for earlier " + incidents + ".");

                Thread delayed_kill = new Thread(new ThreadStart(delegate()
                {
                    int sleep_time = 3;
                    Thread.Sleep(sleep_time * 1000);
                    plugin.KillPlayer(this.name);
                }));

                delayed_kill.Start();
            }

            public List<PlayerProfile> getQueuedPunishersList()
            {
                List<PlayerProfile> list = new List<PlayerProfile>();
                foreach (PlayerPunishment punishment in qpun)
                    list.Add(punishment.punisher);

                return list;
            }

            public int getQueuedPunishmentsCount()
            {
                return qpun.Count;
            }

            public string getQueuedPunishmentsStr(PlayerProfile victim)
            {

                foreach (PlayerPunishment punishment in qpun)
                {
                    if (punishment.punisher != victim)
                        continue;
                    List<string> list = new List<string>();
                    foreach (ViolationType violation in punishment.violations)
                        list.Add(plugin.violation2stringING(violation));

                    return plugin.list2string(list);
                }
                return "";
            }


            public int banVotedBy(PlayerProfile victim)
            {
                if (plugin.isImmune(this) && !plugin.isImmune(victim))
                {
                    plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + "you cannot vote ban " + this.name + " because he has immunity");
                    return 0;
                }

                int violations = getPendingViolations(victim);


                if (violations == 0)
                {
                    //user cannot be vote banned unless a violation exists against the victim
                    plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + "you cannot vote ban  " + this.name + ", unless for " + plugin.getActionableViolationsStr(ActionType.ban));
                    return 0;
                }


                string vstr = getPendingViolationsStr(victim);
                int votes = accumulateV(victim);

                if (this.wasKicked() || this.wasBanned())
                {
                    plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + this.name + " " + plugin.playerstate2stringED(this.state) + " already.");
                    return 0;
                }


                addBanVotesBy(victim, votes);


                int votesNeeded = plugin.getIntegerVarValue("max_ban_votes") - this.getBanVotesReceived() + 1;
                string votesNeededStr = "";
                if (votesNeeded > 1)
                    votesNeededStr = votesNeeded + " ban more votes needed.";
                else if (votesNeeded == 1)
                    votesNeededStr = votesNeeded + " ban more vote needed.";


                if (votes == 1)
                    plugin.SendGlobalMessage(victim.name + " voted to ban " + this.name + " for " + vstr + ". " + votesNeededStr);
                else if (votes > 1)
                    plugin.SendGlobalMessage(victim.name + " cast " + votes + " votes to ban " + this.name + " for " + vstr + ". " + votesNeededStr);


                //Alert other victims to vote
                List<PlayerProfile> other_victims = getVictims();
                foreach (PlayerProfile other_victim in other_victims)
                {
                    if (victim == other_victim)
                        continue;
                    plugin.SendPlayerMessage(other_victim.name, victim.name + " voted to ban " + this.name + ". Type: " + plugin.getActionCommand(ActionType.ban) + " " + this.name.Substring(0, 4));
                }

                return votes;
            }

            public int votedBy(PlayerProfile victim)
            {
                if (plugin.isImmune(this) && !plugin.isImmune(victim))
                {
                    plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + "you cannot vote kick " + this.name + " because he has immunity");
                    return 0;
                }

                int violations = getPendingViolations(victim);

                if (violations == 0)
                {
                    //user cannot be punished unless a violation exists against the victim
                    plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + "you cannot vote against " + this.name + ", unless for " + plugin.getActionableViolationsStr(ActionType.kick));
                    return 0;
                }


                string vstr = getPendingViolationsStr(victim);
                int votes = accumulateV(victim);

                if (this.wasKicked() || this.wasBanned())
                {
                    plugin.SendPlayerMessage(victim.name, plugin.PP(victim.name) + this.name + " " + plugin.playerstate2stringED(this.state) + " already.");
                    return 0;
                }

                addVotesBy(victim, votes);

                int votesNeeded = plugin.getIntegerVarValue("max_kick_votes") - this.getVotesReceived() + 1;
                string votesNeededStr = "";
                if (votesNeeded > 1)
                    votesNeededStr = votesNeeded + " more votes needed.";
                else if (votesNeeded == 1)
                    votesNeededStr = votesNeeded + " more vote needed.";

                if (votes == 1)
                    plugin.SendGlobalMessage(victim.name + " voted to kick " + this.name + " for " + vstr + ". " + votesNeededStr);
                else if (votes > 1)
                    plugin.SendGlobalMessage(victim.name + " cast " + votes + " votes to kick " + this.name + " for " + vstr + ". " + votesNeededStr);


                //Alert other victims to vote
                List<PlayerProfile> other_victims = getVictims();
                foreach (PlayerProfile other_victim in other_victims)
                {
                    if (victim == other_victim)
                        continue;
                    plugin.SendPlayerMessage(other_victim.name, victim.name + " voted to kick " + this.name + ". Type: " + plugin.getActionCommand(ActionType.kick) + " " + this.name.Substring(0, 4));
                }

                return votes;
            }

            private int accumulateV(PlayerProfile victim)
            {
                int violations = 0;

                violations += accumulateTeamKillV(victim);

                return violations;
            }



            private int accumulateTeamKillV(PlayerProfile victim)
            {
                if (tks.ContainsKey(victim))
                {
                    if (!atks.ContainsKey(victim))
                        atks.Add(victim, 0);

                    int v = tks[victim];
                    tks.Remove(victim);
                    atks[victim] += v;

                    return v;
                }

                return 0;
            }


            private int addVotesBy(PlayerProfile victim, int v)
            {
                if (!vtk.ContainsKey(victim))
                    vtk.Add(victim, 0);

                vtk[victim] += v;

                return v;
            }

            private int addBanVotesBy(PlayerProfile victim, int v)
            {
                if (!vtb.ContainsKey(victim))
                    vtb.Add(victim, 0);

                vtb[victim] += v;

                return v;
            }

            public int getPendingViolations(PlayerProfile victim)
            {
                int kills = 0;

                kills += getPendingTeamKills(victim);

                return kills;
            }



            public List<ViolationType> getPendingViolationsList(PlayerProfile victim)
            {
                List<ViolationType> list = new List<ViolationType>();

                if (getPendingTeamKills(victim) > 0)
                    list.Add(ViolationType.team_kill);

                return list;
            }


            public List<ViolationType> getViolationsList()
            {
                List<ViolationType> list = new List<ViolationType>();

                if (getTeamKills() > 0)
                    list.Add(ViolationType.team_kill);

                return list;
            }

            public string getPendingViolationsStr(PlayerProfile victim)
            {

                List<ViolationType> violations = getPendingViolationsList(victim);
                List<string> list = violationsList2StringList(violations);
                return plugin.list2string(list);
            }

            private List<string> violationsList2StringList(List<ViolationType> violations)
            {
                List<string> list = new List<string>();
                foreach (ViolationType violation in violations)
                    list.Add(plugin.violation2stringING(violation));

                return list;
            }

            public string getViolationsStr()
            {
                List<ViolationType> violations = getViolationsList();
                List<string> list = new List<string>();
                foreach (ViolationType violation in violations)
                    list.Add(plugin.violation2stringING(violation));
                return plugin.list2string(list);
            }


            public int getPendingTeamKills(PlayerProfile victim)
            {
                if (tks.ContainsKey(victim))
                    return tks[victim];
                return 0;
            }


            public int getTeamKills(PlayerProfile victim)
            {
                if (atks.ContainsKey(victim))
                    return atks[victim] + getPendingTeamKills(victim);
                return 0;
            }


            public List<PlayerProfile> getTeamKillVictims()
            {
                List<PlayerProfile> victims = new List<PlayerProfile>();
                foreach (KeyValuePair<PlayerProfile, int> pair in tks)
                    if (!victims.Contains(pair.Key))
                        victims.Add(pair.Key);
                foreach (KeyValuePair<PlayerProfile, int> pair in atks)
                    if (!victims.Contains(pair.Key))
                        victims.Add(pair.Key);
                return victims;
            }

            public List<PlayerProfile> getVictims()
            {
                List<PlayerProfile> victims = new List<PlayerProfile>();
                List<PlayerProfile> tk_victims = getTeamKillVictims();


                foreach (PlayerProfile pp in tk_victims)
                    if (!victims.Contains(pp))
                        victims.Add(pp);

                return victims;
            }

            public bool punish(PlayerProfile killer)
            {
                return killer.punishedBy(this, false);
            }

            public bool punish(PlayerProfile killer, Boolean auto)
            {
                return killer.punishedBy(this, auto);
            }

            public bool forgive(PlayerProfile killer)
            {
                return killer.forgivenBy(this);
            }

            public int vote(PlayerProfile myKiller)
            {
                return myKiller.votedBy(this);
            }

            public int ban(PlayerProfile myKiller)
            {
                return myKiller.banVotedBy(this);
            }

            public int getVotesReceived()
            {
                int votes = 0;
                foreach (KeyValuePair<PlayerProfile, int> pair in vtk)
                    votes += pair.Value;

                return votes;
            }

            public int getBanVotesReceived()
            {
                int votes = 0;
                foreach (KeyValuePair<PlayerProfile, int> pair in vtb)
                    votes += pair.Value;

                return votes;
            }

            public int getVotesReceived(PlayerProfile victim)
            {
                if (vtk.ContainsKey(victim))
                    return vtk[victim];
                return 0;
            }

            public int getBanVotesReceived(PlayerProfile victim)
            {
                if (vtb.ContainsKey(victim))
                    return vtb[victim];
                return 0;
            }

            public List<PlayerProfile> getVoters()
            {
                List<PlayerProfile> voters = new List<PlayerProfile>();
                foreach (KeyValuePair<PlayerProfile, int> pair in vtk)
                    if (!voters.Contains(pair.Key))
                        voters.Add(pair.Key);
                return voters;
            }

            public List<PlayerProfile> getBanVoters()
            {
                List<PlayerProfile> voters = new List<PlayerProfile>();
                foreach (KeyValuePair<PlayerProfile, int> pair in vtb)
                    if (!voters.Contains(pair.Key))
                        voters.Add(pair.Key);
                return voters;
            }

            public int getViolations(ViolationType violation)
            {
                if (violation.Equals(ViolationType.team_kill))
                    return getTeamKills();

                return 0;
            }


            public int getTeamKills()
            {
                int tkills = 0;
                foreach (KeyValuePair<PlayerProfile, int> pair in tks)
                    tkills += pair.Value;

                foreach (KeyValuePair<PlayerProfile, int> pair in atks)
                    tkills += pair.Value;

                return tkills;
            }


            public int getViolations()
            {
                return getTeamKills();
            }

            public int getPendingViolations()
            {
                return getPendingTeamKills();
            }


            public int getPendingTeamKills()
            {
                int tkills = 0;
                foreach (KeyValuePair<PlayerProfile, int> pair in tks)
                    tkills += pair.Value;

                return tkills;
            }


            public bool punish()
            {
                if (killer != null)
                    return punish(killer);

                return false;
            }

            public int vote()
            {
                if (killer != null)
                    return vote(killer);

                return 0;
            }

            public int ban()
            {
                if (killer != null)
                    return ban(killer);

                return 0;
            }

            public bool forgive()
            {
                if (killer != null)
                    return forgive(killer);

                return false;
            }

            public bool isAlive()
            {
                return state.Equals(PlayerState.alive);
            }

            public bool isDead()
            {
                return state.Equals(PlayerState.dead) || isViolated();
            }

            public bool wasKicked()
            {
                return state.Equals(PlayerState.kicked);
            }

            public bool wasBanned()
            {
                return state.Equals(PlayerState.banned);
            }

            public bool isViolated()
            {
                return state.Equals(PlayerState.violated);
            }
        }



        public InsanePunisher()
        {


            this.players = new Dictionary<string, PlayerProfile>();

            this.booleanVariables = new Dictionary<string, bool>();



            this.booleanVariables.Add("say_team_kill", false);
            this.booleanVariables.Add("punish_team_kill", false);
            this.booleanVariables.Add("kick_team_kill", false);
            this.booleanVariables.Add("ban_team_kill", false);
            this.booleanVariables.Add("forgive_team_kill", false);
            this.booleanVariables.Add("auto_punish_team_kill", false);


            this.booleanVariables.Add("mock_suicide", true);
            this.booleanVariables.Add("insult_player", true);
            this.booleanVariables.Add("slap_player", true);
            this.booleanVariables.Add("quiet_mode", false);

            this.integerVariables = new Dictionary<string, int>();
            this.integerVariables.Add("max_ban_votes", 3);
            this.integerVariables.Add("ban_minutes", 20);
            this.integerVariables.Add("max_kick_votes", 3);
            this.integerVariables.Add("auto_action_max", 3);
            this.integerVariables.Add("debug_level", 3);


            this.floatVariables = new Dictionary<string, float>();

            this.stringListVariables = new Dictionary<string, string>();
            this.stringListVariables.Add("admin_list", @"micovery|admin2|admin3");
            this.stringListVariables.Add("player_white_list", @"micovery|player2|player3");
            this.stringListVariables.Add("mock_msg_list", String.Join("|", this.default_mocks.ToArray()));
            this.stringListVariables.Add("insult_msg_list", String.Join("|", this.default_insults.ToArray()));
            this.stringListVariables.Add("slap_msg_list", String.Join("|", this.default_slaps.ToArray()));
            this.stringListVariables.Add("team_kill_warn_list", String.Join("|", this.default_warn_team_kill.ToArray()));



            this.stringVariables = new Dictionary<string, string>();
            this.stringVariables.Add("console", "Type a command here to test");

            this.enumVariables = new Dictionary<string, string>();
            this.enumVariablesType = new Dictionary<string, Type>();
            this.enumVariables.Add("ban_type", BanType.GUID.ToString());
            this.enumVariablesType.Add("ban_type", typeof(BanType));
            this.enumVariables.Add("ban_duration", BanDuration.Temporary.ToString());
            this.enumVariablesType.Add("ban_duration", typeof(BanDuration));
            this.enumVariables.Add("auto_action", AutoAction.None.ToString());
            this.enumVariablesType.Add("auto_action", typeof(AutoAction));



            this.hiddenVariables = new List<string>();
            this.hiddenVariables.Add("auto_punish_team_kill");

            this.hiddenVariables.Add("team_kill_warn_list");

            this.hiddenVariables.Add("mock_msg_list");
            this.hiddenVariables.Add("insult_msg_list");
            this.hiddenVariables.Add("slap_msg_list");
            this.hiddenVariables.Add("mock_msg_list");
            this.hiddenVariables.Add("map_list");


            /* create variable groups */
            variableGroups = new Dictionary<string, List<string>>();


            variableGroups.Add("Team Killing", new List<string>());
            variableGroups["Team Killing"].Add("say_team_kill");
            variableGroups["Team Killing"].Add("punish_team_kill");
            variableGroups["Team Killing"].Add("kick_team_kill");
            variableGroups["Team Killing"].Add("ban_team_kill");
            variableGroups["Team Killing"].Add("forgive_team_kill");

            variableGroups.Add("Whitelists", new List<string>());
            variableGroups["Whitelists"].Add("player_white_list");

            variableGroups.Add("Voting", new List<string>());
            variableGroups["Voting"].Add("max_ban_votes");
            variableGroups["Voting"].Add("max_kick_votes");

            variableGroups.Add("Banning", new List<string>());
            variableGroups["Banning"].Add("ban_type");
            variableGroups["Banning"].Add("ban_duration");
            variableGroups["Banning"].Add("ban_minutes");

            variableGroups.Add("Messages", new List<string>());
            variableGroups["Messages"].Add("mock_suicide");
            variableGroups["Messages"].Add("slap_player");
            variableGroups["Messages"].Add("insult_player");

            variableGroups.Add("Automation", new List<string>());
            variableGroups["Automation"].Add("auto_action");
            variableGroups["Automation"].Add("auto_action_max");
            variableGroups["Automation"].Add("auto_punish_team_kill");


            variableGroups.Add("Advanced", new List<string>());
            variableGroups["Advanced"].Add("mock_msg_list");
            variableGroups["Advanced"].Add("slap_msg_list");
            variableGroups["Advanced"].Add("insult_msg_list");
            variableGroups["Advanced"].Add("team_kill_warn_list");


            /* Set the order of the groups */
            variableGroupsOrder = new List<string>();
            variableGroupsOrder.Add("Settings");
            variableGroupsOrder.Add("Team Killing");
            variableGroupsOrder.Add("Voting");
            variableGroupsOrder.Add("Banning");
            variableGroupsOrder.Add("Automation");
            variableGroupsOrder.Add("Messages");
            variableGroupsOrder.Add("Whitelists");
            variableGroupsOrder.Add("Advanced");

        }


        public void loadSettings()
        {


        }


        public void unloadSettings()
        {
            this.players.Clear();
        }



        public string GetPluginName()
        {
            return "Insane Punisher";
        }

        public string GetPluginVersion()
        {
            return "0.0.0.1-patch-2 (beta, BF3)";
        }

        public string GetPluginAuthor()
        {
            return "micovery";
        }

        public string GetPluginWebsite()
        {
            return "www.insanegamersasylum.com";
        }


        public string GetPluginDescription()
        {
            return @"
        <h2>Description</h2>
        <p>This is a re-write of the famous Insane Punisher for Battlefield Bad Company 2. In this version I have removed all the features not supported by Battlefield 3.<br />
        <br />
        <h2>Features Summary</h2>
        <br />
        <ul>
          <li><b>Retribution</b><br />
          <br />
          Victims of team-killers are allowed to !punish, !forgive, !votekick, and !voteban <br />
          <br />
          </li>
          <li><b>Queued Punishments</b><br />
          <br />
          If the team-killer is dead at the time of !punish, the punishment is queued, and enfornced the next time he spawns.<br />
          <br />
          </li>
          <li><b>Individual Punishments</b><br />
          <br />
          If victim was team-killed by multiple players, the plugin remembers all the team-killers. This allows the victim to punish the team-killer, using the <b>!punish {player}</b> command, even if <b>{player}</b> was not the last one to team-kill.<br />
          <br />
          </i>
          <li><b>White-list</b><br />
          <br />
          Players that are in the white-list have immunity against all punishments.<br />
          <br />
          </li>
          <li><b>Automatic Punishments</b><br />
          <br />
          If desired, you can set the plugin so that it automatically punishes the team-killer immediately after each violation<br />
          <br />
          </li>
          <li><b>Automatic Actions</b><br />
          <br />
          If desired, you can set the plugin to automatically kick or ban team-killers after they receive a certain number of punishments<br />
          <br />
          </li>
          <li><b>Customizable Warnings</b><br /> 
          <br />
          Allows you to change the default team-kill warnings that are sent to the team-killers <br />
          <br />
          </li>
          <li><b>Customizable Slaps</b><br /> 
          <br />
          Allows you to have a !slap command for players to have fun and bitch slap each-other<br />
          <br />
          </li>
          <li><b>Customizable Insults</b><br />
          <br />
          Allows you to have an !insult command for players to have fun and insult each-other<br />
          <br />
          <li><b>Track Violations</b><br />
          <br />
          Allows you to use commands such as <b>!violations</b> or <b>!violations {player}</b> to view a detailed history of player violations during the round<br />
          <br />
          </li>
        </ul>
        <br />
        <h2>Missing Features</h2>
        <br />
        <ul>
          <li>Punishment for Base-Raping </li>
          <li>Punishment for Base-Camping</li>
          <li>Private messages</i>
          <li>White-list based on Clan-Tag</li>
        </ul>

        </p>


        <h2>Rules of Engagement - Punishable Incidents</h2>
        
        <pre>    
            Legend
            --  base boundary
            --  defense boundary
  

                  Team Killing
            (inside or outside base)  
           +-------------------------+
                                    
                 o      \o_         
                 \=-->  /           
                 /\   //            
                                        o      \o_
                                        \=-->  /
                                        /\   //
                                    
           +-------------------------+

                                     
        </pre>  


        <h2>Settings</h2>
        <ol>
          <li><blockquote><strong>console</strong><br />
                <i>(String)</i> - used for running commands from within ProCon<br />
                <br />
                e.g !violations micovery<br />
                <br />
                </blockquote> 
          </li>
          <li><blockquote><strong>debug_level</strong><br />
                <i>(integer >= 0)</i> - Level of tracing messages printed to the plugin console. 0 is the lowest, with least debug messages. The higher the value, the more tracing messages could be printed. <br />
                </blockquote> 
          </li>
          <li><blockquote><strong>max_kick_votes</strong><br />
                <i>(integer > 0)</i> - maximum number of votes that a soldier can receive before being kicked. <br />
                e.g if max_kick_votes = 3, soldier is kicked that the 4th vote.  <br />
                </blockquote> 
          </li> 
          <li><blockquote><strong>max_ban_votes</strong><br />
                <i>(integer > 0)</i> - maximum number of votes that a soldier can receive before being banned. <br />
                e.g if max_ban_votes = 3, soldier is banned that the 4th vote.  <br />
                </blockquote> 
          </li> 
          <li><blockquote> <strong>say_team_kill</strong><br />
                <i>true</i> - inform killer and victim (in chat box) of team kill violation <br />
                <i>false</i> - do not inform killer or victim of team kill violation  
                </blockquote> 
          </li>
           <li><blockquote><strong>punish_team_kill</strong><br />
                <i>true</i> - allows victim to punish the team killer <br />
                <i>false</i> - forbids victim from punishing the team killer
                </blockquote> 
           </li>
           <li><blockquote><strong>forgive_team_kill</strong><br />
                <i>true</i> - allows victim to forgive the team killer <br />
                <i>false</i> - forbids victim from forgiving the team killer
                </blockquote> 
           </li>
           <li><blockquote><strong>kick_team_kill</strong><br />
                <i>true</i> - allows victim to vote kick the team killer <br />
                <i>false</i> - forbids victim from vote kicking the team killer
                </blockquote> 
           </li>
           <li><blockquote><strong>ban_team_kill</strong><br />
                <i>true</i> - allows victim to vote ban the team killer <br />
                <i>false</i> - forbids victim from vote banning the team killer
                </blockquote> 
           </li>
           <li><blockquote><strong>mock_suicide</strong><br />
                <i>true</i> - server mocks the soldier that suicided (private message) <br />
                <i>false</i> - no message sent on suicide
                </blockquote> 
           </li>
           <li><blockquote><strong>slap_player</strong><br />
                <i>true</i> - allows soldiers to slap each other with a big smelly tuna fish <br />
                <i>false</i> - forbids soldiers from slapping each other 
                </blockquote> 
           </li>
           <li><blockquote><strong>insult_player</strong><br />
                <i>true</i> - allows soldiers to insult each other, random insult is sent privately<br />
                <i>false</i> - forbids soldiers from insulting each other
                </blockquote> 
           </li>
           <li><blockquote><strong>admin_list</strong><br />
                <i>(Stringlist)</i> - list of soldiers that are allowed to modify/view plugin settings while in-game.<br />
                </blockquote> 
           </li>
           <li><blockquote><strong>player_white_list</strong><br />
                <i>(Stringlist)</i> - list of soldiers that have immunity against punish, kick, and ban <br />
                </blockquote> 
           </li>
           <li><blockquote><strong>ban_type</strong><br />
                <i>GUID</i> - ban players by EA GUID <br />
                <i>Name</i> - ban players by EA soldier name<br />
                <i>IPAddress</i> - ban players by IP Address <br />
                </blockquote> 
           </li> 
           <li><blockquote><strong>ban_duration</strong><br />
                <i>Permanent</i> - ban indefinitely <br />
                <i>Round</i> - ban until the end of the current round<br />
                <i>Temporary</i> - ban temporarily (you need to set <b>ban_minutes</b> as well)<br />
                </blockquote> 
           </li>
          <li><blockquote><strong>ban_minutes</strong><br />
                <i>(integer > 0)</i> - Number of minutes to ban a player after being vote banned <br />
                </blockquote> 
          </li>
         <!--
          <li><blockquote><strong>advanced_mode</strong><br />
                <i>true</i> - shows the advanced settings group <br />
                <i>false</i> - hides the advanced settings group 
                </blockquote> 
          </li>
         -->        
        </ol>
        <h2>Advanced Settings</h2>
        <p>
           Do not use these settings if you are just beginning to learn how to use this plugin. These settings allow you to tweak default behaviors for actions againts rule violators, as well as default messages.
        </p>
        <ol>

          <li><blockquote><strong>auto_action</strong><br />
              <i>None</i> - no action is perfomed when <b>auto_action_max</b> threshold is reached<br />
              <i>Ban</i> - player is banned when <b>auto_action_max</b> threshold is reached <br />
              <i>Kick</i> - player is kicked when <b>auto_action_max</b> threshold is reached  <br />
              </blockquote> 
          </li>
          <li><blockquote><strong>auto_action_max</strong><br />
              <i>(interger >= 1)</i> -  maximum number of punishments that a player can receive before being auto kicked or banned<br />
              For example, if the value of <b>auto_action_max</b> is 3, the player will be kicked/banned on the 4th punishment received<br />
              </blockquote> 
          </li>
          <li><blockquote><strong>auto_punish_team_kill</strong><br />
              <i>true</i> - team killer is automatically punished, and victim is not allowed to take any action <br />
              <i>false</i> - team killer is not automatically punished <br />
              </blockquote> 
          </li>
          Note that when <b>auto_punish_*</b> settings are enabled, then the corresponding setting <b>punish_*</b> is ignored.
          <li><blockquote><strong>slap_msg_list</strong><br />
              <i>(StringList)</i> - list of messages used for the <b>!slap</b> command  <br />
              </blockquote> 
          </li>
          <li><blockquote><strong>mock_msg_list</strong><br />
              <i>(StringList)</i> - list of messages used to mock a player after suicide  <br />
              </blockquote> 
          </li>
          <li><blockquote><strong>insult_msg_list</strong><br />
              <i>(StringList)</i> - list of messages used for the <b>!insult</b> command  <br />
              </blockquote> 
          </li>
          <li><blockquote><strong>team_kill_warn_list</strong><br />
              <i>(StringList)</i> - ordered list of warning messages that are sent to team killers after a violation. <br />
              </blockquote> 
          </li>
          <p>
              Note that the order of the messages in the warning lists is relevant. <br />
              The first message is sent when the first violation happends. <br />
              The second message is sent when the second violation happens, and so on.
          </p>      
        </ol>    

        <h2>Public Commands</h2>
        <p>
            Commands are messages typed into the game chat box, or into the <b>console</b> field of the plugin.
            Commands must start with one of the following characters: !,@,?, or /. This plugin interprets the following public commands.
        </p>
        <ul>
           <li><blockquote><strong>!punish [name-substring]</strong><br />
               This is the main command, and most fun of all. It allows victims to punish soldiers who violated the rules of engagement. 
               If ""!punish"" is used without specifying the soldier name, then the last aggressor with ROE violation (against the victim) is punished.
               Punishments cannot be stacked. For example, if an aggressor team kills a victim twice, then the victim can only punish the aggressor once. 
               However, punishments by different victims are stackable. For example, if an aggressor team kills two victims, and both decide to punish him, 
               then the aggressor is punished twice consecutively. Note that if an aggressor is dead, punishments are queued and applied on re-spawn.
               </blockquote> 
           </li> 
           <li><blockquote> <strong>!votekick [name-substring]</strong><br />
               This is a lesser used command, but still powerful. It allows victims to cast votes against soldiers who violated the rules of engagement.
               If ""!votekick"" is used without specifying the soldier name, then the last aggressor with ROE violation (against the victim) is voted against.
               Votes are stackable. For example, if an aggressor team kills a victim twice, he can vote once against the aggressor, with two votes being cast.
               When a victim casts a kick vote against an aggressor, all other victims of that aggressor, who have not taken action, are reminded to vote.
               This is similar to a traditional votekick, with the difference that only victims of the aggressor are allowed to vote.
            </blockquote> 
           <li><blockquote> <strong>!voteban [name-substring]</strong><br />
               This is also a lesser used command, but extremely powerful. It allows victims to cast votes against soldiers who violated the rules of engagement.
               If ""!voteban"" is used without specifying the soldier name, then the last aggressor with ROE violation (against the victim) is voted against.
               Votes are stackable. For example, if an aggressor team kills a victim twice, he can vote once against the aggressor, with two votes being cast.
               When a victim casts a ban vote against an aggressor, all other victims of that aggressor, who have not taken action, are reminded to vote.
               This is similar to a traditional voteban, with the difference that only victims of the aggressor are allowed to vote.
            </blockquote> 
           </li>
           <li><blockquote> <strong>!forgive [name-substring]</strong><br />
               This is a command of camaraderie. It allows victims to forgive soldiers who violated the rules of engagement. When forgiving, the victim waives
               the right to take action against the aggressor for all previous ROE violatons. This command is mostly used by soldiers forgiving teammates afer an 
               accidental team kill. 
            </blockquote> 
           <li><blockquote> <strong>!slap [name-substring]</strong><br />
               This command is just for fun. It allows soldiers to slap each other publicly. The type of slap is chosen randomly from the settings file. 
            </blockquote> 
           <li><blockquote><strong>!insult [name-substring]</strong><br />
               This is another useless, but fun command. It allows soldiers to privately insult each other. The pre-defined insults in the settings file are mostly
               jokingly insults. Insults are chosen at random. 
            </blockquote> 
           <li><blockquote><strong>!my violations</strong><br />
               This is a statistics command. It allows a soldier to view his rules of engagement violations, such as team killing, base raping, and base camping.
               It displays the name of the victims, together with the total and pending violations against each victim. Pending violations are those
               for which the victim has not yet taken action.
            </blockquote> 
           </li>
        </ul>
       <h2> Admin Commands</h2>
        <p>
            These are the commands that only players in the ""admin_list"" are allowed to execute. Reply messages generated by these commands are sent only to the squad of the admin who executed the command.
        </p>
        <ul>
           <li><blockquote><strong>!punisher settings</strong><br />
                This command prints the values of all plugin variables.
               </blockquote> 
           </li>
           <li><blockquote>
                <strong>1. !set {variable} {to|=} {value}</strong><br />
                <strong>2. !set {variable} {value}</strong><br />       
                <strong>3. !set {variable}</strong><br />   
                This command is used for setting the value of this plugin's variables.<br />
                For the 3rd invocation syntax the value is assumed to be ""true"".
               </blockquote> 
           </li>
           <li><blockquote>
                <strong>!get {variable} </strong><br />
                This command prints the value of the specified variable.
               </blockquote> 
           </li>
           <li><blockquote>
                <strong>!enable {variable-substring}</strong><br />
                This command is a shortcut for enabling  (setting to ""true"") multiple variables at the same time.<br />
                For example, the following command: <br />
                <br />
                !enable punish <br />
                <br />
                results in all variables that contain the word ""punish"" being set to ""true""<br />
               </blockquote> 
           </li>
           <li><blockquote>
                <strong>!disable {variable-substring}</strong><br />
                This command is a shortcut for disabling  (setting to ""false"") multiple variables at the same time.<br />
                For example, the following command: <br />
                <br />
                !disable punish <br />
                <br />
                results in all variables that contain the word ""punish"" being set to ""false""<br /> 
               </blockquote> 
           </li>
           <li><blockquote>
                <strong>!violations {name-substring}</strong><br />
                This command prints a detailed report of the specified soldier's ROE violations. This is similar to !my violations command. <br />
               </blockquote> 
           </li>
           <li><blockquote>
                <strong>!violations</strong><br />
                This command prints a summary of all soldiers' ROE violations<br />
               </blockquote> 
           </li>
         </ul> 
        ";
        }

        public void OnPluginLoaded(string strHostName, string strPort, string strPRoConVersion)
        {
            ConsoleWrite("plugin loaded");
        }

        public void OnPluginEnable()
        {
            ConsoleWrite("^b^2Enabled!^0");

            unloadSettings();
            loadSettings();

            getPlayersList();
            getPBPlayersList();
        }


        public void OnPluginDisable()
        {

            ConsoleWrite("^1^bDisabled =(^0");

            unloadSettings();
        }

        public List<CPluginVariable> GetDisplayPluginVariables()
        {
            List<CPluginVariable> lstReturn = new List<CPluginVariable>();




            List<string> vars = getPluginVars();
            foreach (string var in vars)
            {
                int temp;
                if (Regex.Match(var, @"^(quiet_mode)$").Success)
                    continue;

                string group = getPluginVarGroup(var);
                string type = "multiline";
                string value = getPluginVarValue(var, true);

                /* build boolean enum */
                if (isBooleanVar(var))
                    type = getEnumStr(var, typeof(BooleanTrueFalse));
                /* build the enum type for enums */
                else if (isEnumVar(var))
                    type = getEnumStr(var, getEnumVarType(var));

                if (variableGroupsOrder.Contains(group))
                    group = (variableGroupsOrder.IndexOf(group) + 1).ToString() + ". " + group;

                if (isStringListVar(var))
                    type = "stringarray";

                lstReturn.Add(new CPluginVariable(group + "|" + var, type, value));
            }

            return lstReturn;
        }


        public string getEnumStr(string var, Type type)
        {
            if (!type.IsEnum)
                return "";
            return "enum." + var + "(" + String.Join("|", Enum.GetNames(type)) + ")";
        }

        // Lists all of the plugin variables.
        /*public List<CPluginVariable> GetPluginVariables()
        {
            List<CPluginVariable> lstReturn = new List<CPluginVariable>();

            List<string> vars = getPluginVars();
            foreach (string var in vars)
                lstReturn.Add(new CPluginVariable(var, typeof(string), getPluginVarValue(var)));

            return lstReturn;
        }*/

        public List<CPluginVariable> GetPluginVariables()
        {
            List<CPluginVariable> lstReturn = new List<CPluginVariable>();

            List<string> vars = getPluginVars();
            foreach (string var in vars)
                lstReturn.Add(new CPluginVariable(var, typeof(string), "BASE64:" + Encode(getPluginVarValue(var))));

            return lstReturn;
        }


        /*
        public void SetPluginVariable(string strVariable, string strValue)
        {
            //ConsoleWrite("setting " + strVariable + " to " + strValue);
            if (strVariable.ToLower().Contains("refresh"))
                return;

            setPluginVarValue(strVariable, strValue);
        }*/

        public void SetPluginVariable(string var, string val)
        {
            try
            {
                String decoded = val;
                bool ui = true;
                if (decoded.StartsWith("BASE64:"))
                {
                    decoded = decoded.Replace("BASE64:", "");
                    decoded = Decode(decoded);
                    ui = false;
                }

                setPluginVarValue(var, decoded, ui);
            }
            catch (Exception e)
            {
                DumpException(e);
            }
        }



        // Player events
        public override void OnPlayerJoin(string strSoldierName)
        {


        }


        public void PluginCommand(String sender, String text)
        {
            try
            {
                if (IsInGameCommand(text))
                    InGameCommand(sender, text);
            }
            catch (Exception e)
            {
                DumpException(e);
            }
        }

        // Will receive ALL chat global/team/squad in R3.
        public override void OnGlobalChat(string sender, string text)
        {
            PluginCommand(sender, text);
        }

        // Place holder, non-functioning in R3.  It recieves the same data as OnGlobalChat though so look out for now.
        public override void OnTeamChat(string sender, string text, int iTeamID)
        {
            PluginCommand(sender, text);
        }

        // Place holder, non-functioning in R3.  It recieves the same data as OnGlobalChat though so look out for now.
        public override void OnSquadChat(string sender, string text, int iTeamID, int iSquadID)
        {
            PluginCommand(sender, text);
        }





        public override void OnPunkbusterPlayerInfo(CPunkbusterInfo cpbiPlayer)
        {
            try
            {
                if (cpbiPlayer == null)
                    return;

                processNewPlayer(cpbiPlayer);
            }
            catch (Exception e)
            {
                DumpException(e);
            }
        }


        public void processNewPlayer(CPunkbusterInfo cpbiPlayer)
        {
            if (this.players.ContainsKey(cpbiPlayer.SoldierName))
                this.players[cpbiPlayer.SoldierName].pbinfo = cpbiPlayer;
            else
            {
                lock (mutex)
                {
                    players.Add(cpbiPlayer.SoldierName, new PlayerProfile(this, cpbiPlayer));
                }
            }
        }




        public override void OnPlayerKilled(Kill killInfo)
        {

            try
            {

                //get the killer and victim information

                CPlayerInfo killer = killInfo.Killer;
                CPlayerInfo victim = killInfo.Victim;

                PlayerProfile victimProfile = getPlayerProfile(victim.SoldierName);
                PlayerProfile killerProfile = getPlayerProfile(killer.SoldierName);



                //update the victim state
                if (killerProfile != null && victimProfile != null && !victimProfile.isViolated())
                    victimProfile.killedBy(killerProfile);


                if (killInfo.IsSuicide)
                {
                    mockPlayerSuicide(victim.SoldierName);
                    victimProfile.state = PlayerState.dead;
                    return;
                }
                else if (killer.TeamID == victim.TeamID)
                {
                    TeamKill(killer, victim);
                }
            }
            catch (Exception e)
            {
                DumpException(e);
            }
        }

        public override void OnRoundOver(int iWinningTeamID)
        {
            try
            {
                //reset all player statistics
                foreach (KeyValuePair<string, PlayerProfile> pair in this.players)
                {
                    PlayerProfile player = pair.Value;
                    player.resetStats();
                }
            }
            catch (Exception e)
            {
                DumpException(e);
            }
        }

        public override void OnPlayerSpawned(string soldierName, Inventory spawnedInventory)
        {
            try
            {
                PlayerProfile playerProfile = getPlayerProfile(soldierName);
                if (playerProfile != null)
                {
                    playerProfile.state = PlayerState.alive;

                    playerProfile.dequeueMessages();
                    playerProfile.dequeuePunishment();
                }
            }
            catch (Exception e)
            {
                DumpException(e);
            }
        }


        public override void OnLoadingLevel(string mapFileName, int roundsPlayed, int roundsTotal)
        {

        }


        private void KillPlayerWithMessage(string soldierName, string message)
        {
            SendPlayerMessage(soldierName, message);
            KillPlayer(soldierName);
        }

        private void KillPlayer(string soldierName)
        {
            PlayerProfile player = getPlayerProfile(soldierName);

            if (player != null && !player.isViolated())
                player.state = PlayerState.dead;

            this.ExecuteCommand("procon.protected.send", "admin.killPlayer", soldierName);
        }


        private void KickPlayerWithMessage(string soldierName, string message)
        {

            PlayerProfile player = getPlayerProfile(soldierName);

            if (player != null)
            {
                player.state = PlayerState.kicked;
                this.ExecuteCommand("procon.protected.send", "admin.kickPlayer", player.name, message);
            }
        }

        private void BanPlayerWithMessage(BanType type, BanDuration duration, String soldierName, int minutes, string message)
        {

            PlayerProfile player = getPlayerProfile(soldierName);
            if (player == null)
                return;

            /* get the type field and value */
            string typeField = "guid";
            string typeValue = player.info.GUID;

            if (type.Equals(BanType.GUID))
            {
                typeField = "guid";
                typeValue = player.info.GUID;
            }
            else if (type.Equals(BanType.IPAddress))
            {
                typeField = "ip";
                typeValue = player.pbinfo.Ip.Remove(player.pbinfo.Ip.IndexOf(':'));
            }
            else if (type.Equals(BanType.Name))
            {
                typeField = "name";
                typeValue = player.info.SoldierName;
            }

            /* get the time out value */
            string timeout = "seconds";
            if (duration.Equals(BanDuration.Permanent))
                timeout = "perm";
            else if (duration.Equals(BanDuration.Round))
                timeout = "round";
            else if (duration.Equals(BanDuration.Temporary))
                timeout = "seconds";

            if (duration.Equals(BanDuration.Temporary))
                message += ", Temporary-Ban " + minutes + " minutes";
            else if (duration.Equals(BanDuration.Round))
                message += ", Round-Ban ";
            else if (duration.Equals(BanDuration.Permanent))
                message += ", Permanent-Ban";


            // truncate the string if it exceeds 80 characters
            if (message.Length > 80)
                message = message.Substring(0, 80);


            player.state = PlayerState.banned;
            if (duration.Equals(BanDuration.Temporary))
                this.ExecuteCommand("procon.protected.send", "banList.add", typeField, typeValue, timeout, (minutes * 60).ToString(), message);
            else
                this.ExecuteCommand("procon.protected.send", "banList.add", typeField, typeValue, timeout, message);

            this.ExecuteCommand("procon.protected.send", "banList.save");
            KickPlayerWithMessage(soldierName, message);
        }


        private void KickPlayer(string soldierName)
        {

            PlayerProfile player = getPlayerProfile(soldierName);

            if (player != null)
            {
                player.state = PlayerState.kicked;
                this.ExecuteCommand("procon.protected.send", "admin.kickPlayer", player.name, "");
            }
        }

        public void ServerCommand(params string[] args)
        {
            List<string> list = new List<string>();
            list.Add("procon.protected.send");
            list.AddRange(args);
            this.ExecuteCommand(list.ToArray());
        }

        private bool SendSquadMessage(int teamId, int squadId, String message)
        {
            ServerCommand("admin.say", message, "squad", (teamId).ToString(), (squadId).ToString());
            return true;
        }

        private void SendPlayerMessage(string name, string message)
        {

            if (getBooleanVarValue("quiet_mode") && !isAdmin(name))
                return;

            // special case for handling commands executed from the "console" field
            if (name == null)
            {
                ConsoleWrite(message);
                return;
            }

            PlayerProfile pp = getPlayerProfile(name);
            if (pp == null || pp.info.TeamID == -1 || pp.info.SquadID == -1)
                return;

            SendSquadMessage(pp.info.TeamID, pp.info.SquadID, message);
        }

        private void SendGlobalMessage(string message)
        {
            if (getBooleanVarValue("quiet_mode"))
                DebugWrite(message, 3);
            else
                ServerCommand("admin.say", message, "all");

        }

        private void SendConsoleMessage(string name, string message)
        {
            if (name != null)
                SendPlayerMessage(name, message);

            ConsoleWrite(message);
        }

        private void SendConsoleWarn(string name, string message)
        {
            if (name != null)
                SendPlayerMessage(name, message);

            ConsoleWarn(message);
        }

        public void ConsoleWrite(string msg, MessageType type)
        {
            LogWrite(FormatMessage(msg, type));
        }

        public void ConsoleWrite(string msg)
        {
            ConsoleWrite(msg, MessageType.Normal);
        }

        public void ConsoleWarn(String msg)
        {
            ConsoleWrite(msg, MessageType.Warning);
        }

        public void ConsoleError(String msg)
        {
            ConsoleWrite(msg, MessageType.Error);
        }

        public void ConsoleException(String msg)
        {
            ConsoleWrite(msg, MessageType.Exception);
        }

        public void DebugWrite(string msg, int level)
        {
            if (getIntegerVarValue("debug_level") >= level)
                ConsoleWrite(msg, MessageType.Normal);
        }


        public void LogWrite(String msg)
        {
            this.ExecuteCommand("procon.protected.pluginconsole.write", msg);
        }

        public enum MessageType { Warning, Error, Exception, Normal };

        public String FormatMessage(String msg, MessageType type)
        {
            String prefix = "[^b" + GetPluginName() + "^n] ";

            if (Thread.CurrentThread.Name != null)
                prefix += "Thread(^b" + Thread.CurrentThread.Name + "^n): ";

            if (type.Equals(MessageType.Warning))
                prefix += "^1^bWARNING^0^n: ";
            else if (type.Equals(MessageType.Error))
                prefix += "^1^bERROR^0^n: ";
            else if (type.Equals(MessageType.Exception))
                prefix += "^1^bEXCEPTION^0^n: ";

            return prefix + msg;
        }


        public static String InGameCommand_Pattern = @"^\s*[@/!\?]\s*";

        public bool IsInGameCommand(String text)
        {
            return Regex.Match(text, InGameCommand_Pattern).Success;
        }

        public String ExtractInGameCommand(String text)
        {
            return Regex.Replace(text, InGameCommand_Pattern, "").Trim();
        }

        private void InGameCommand(string sender, string cmd)
        {


            cmd = ExtractInGameCommand(cmd);

            //Setting/Getting variables
            Match setVarValueMatch = Regex.Match(cmd, @"^set\s+([^ ]+)\s+(.+)", RegexOptions.IgnoreCase);
            Match setVarValueEqMatch = Regex.Match(cmd, @"^set\s+([^ ]+)\s*=\s*(.+)", RegexOptions.IgnoreCase);
            Match setVarValueToMatch = Regex.Match(cmd, @"^set\s+([^ ]+)\s+to\s+(.+)", RegexOptions.IgnoreCase);
            Match setVarTrueMatch = Regex.Match(cmd, @"^set\s+([^ ]+)", RegexOptions.IgnoreCase);
            Match getVarValueMatch = Regex.Match(cmd, @"^get\s+([^ ]+)", RegexOptions.IgnoreCase);
            Match enableMatch = Regex.Match(cmd, @"^enable\s+(.+)", RegexOptions.IgnoreCase);
            Match disableMatch = Regex.Match(cmd, @"^disable\s+(.+)", RegexOptions.IgnoreCase);


            //Player actions2
            Match punishMatch = Regex.Match(cmd, @"^punish", RegexOptions.IgnoreCase);
            Match punishPlayerMatch = Regex.Match(cmd, @"^punish\s+(.+)", RegexOptions.IgnoreCase);
            Match voteMatch = Regex.Match(cmd, @"^votekick", RegexOptions.IgnoreCase);
            Match votePlayerMatch = Regex.Match(cmd, @"^votekick\s+(.+)", RegexOptions.IgnoreCase);
            Match forgiveMatch = Regex.Match(cmd, @"^forgive", RegexOptions.IgnoreCase);
            Match forgivePlayerMatch = Regex.Match(cmd, @"^forgive\s+(.+)", RegexOptions.IgnoreCase);
            Match banMatch = Regex.Match(cmd, @"^voteban", RegexOptions.IgnoreCase);
            Match banPlayerMatch = Regex.Match(cmd, @"^voteban\s+(.+)", RegexOptions.IgnoreCase);


            //Information
            Match playerStatsMatch = Regex.Match(cmd, @"^violations\s+([^ ].*)", RegexOptions.IgnoreCase);
            Match playersStatsMatch = Regex.Match(cmd, @"^violations", RegexOptions.IgnoreCase);
            Match pluginSettingsMatch = Regex.Match(cmd, @"^punisher(\s*settings)?", RegexOptions.IgnoreCase);

            Match myStatsMatch = Regex.Match(cmd, @"^my\s+violations", RegexOptions.IgnoreCase);
            Match slapPlayerMatch = Regex.Match(cmd, @"^slap\s+([^ ]+)", RegexOptions.IgnoreCase);
            Match insultPlayerMatch = Regex.Match(cmd, @"^insult\s+([^ ]+)", RegexOptions.IgnoreCase);


            bool senderIsAdmin = (sender == null) ? true : isAdmin(sender);


            if (myStatsMatch.Success)
                myStatsCmd(sender);
            else if (playerStatsMatch.Success && senderIsAdmin)
                playerStatsCmd(sender, playerStatsMatch.Groups[1].Value);
            else if (playersStatsMatch.Success && senderIsAdmin)
                playersStatsCmd(sender);
            else if (setVarValueEqMatch.Success && senderIsAdmin)
                setVariableCmd(sender, setVarValueEqMatch.Groups[1].Value, setVarValueEqMatch.Groups[2].Value);
            else if (setVarValueToMatch.Success && senderIsAdmin)
                setVariableCmd(sender, setVarValueToMatch.Groups[1].Value, setVarValueToMatch.Groups[2].Value);
            else if (setVarValueMatch.Success && senderIsAdmin)
                setVariableCmd(sender, setVarValueMatch.Groups[1].Value, setVarValueMatch.Groups[2].Value);
            else if (setVarTrueMatch.Success && senderIsAdmin)
                setVariableCmd(sender, setVarTrueMatch.Groups[1].Value, "1");
            else if (getVarValueMatch.Success && senderIsAdmin)
                getVariableCmd(sender, getVarValueMatch.Groups[1].Value);
            else if (enableMatch.Success && senderIsAdmin)
                enableVarGroupCmd(sender, enableMatch.Groups[1].Value);
            else if (disableMatch.Success && senderIsAdmin)
                disableVarGroupCmd(sender, disableMatch.Groups[1].Value);
            else if (pluginSettingsMatch.Success && senderIsAdmin)
                pluginSettingsCmd(sender);
            else if (punishPlayerMatch.Success)
                punishMyKillerCmd(sender, punishPlayerMatch.Groups[1].Value);
            else if (punishMatch.Success)
                punishMyKillerCmd(sender);
            else if (votePlayerMatch.Success)
                voteMyKillerCmd(sender, votePlayerMatch.Groups[1].Value);
            else if (voteMatch.Success)
                voteMyKillerCmd(sender);
            else if (forgivePlayerMatch.Success)
                forgiveMyKillerCmd(sender, forgivePlayerMatch.Groups[1].Value);
            else if (forgiveMatch.Success)
                forgiveMyKillerCmd(sender);
            else if (banPlayerMatch.Success)
                banMyKillerCmd(sender, banPlayerMatch.Groups[1].Value);
            else if (banMatch.Success)
                banMyKillerCmd(sender);
            else if (slapPlayerMatch.Success)
                slapPlayerCmd(sender, slapPlayerMatch.Groups[1].Value);
            else if (insultPlayerMatch.Success)
                insultPlayerCmd(sender, insultPlayerMatch.Groups[1].Value);

        }

        private void slapPlayerCmd(string sender, string victim)
        {

            List<String> slap_list = getStringListVarValue("slap_msg_list");

            if (getBooleanVarValue("slap_player") == false)
            {
                SendPlayerMessage(sender, PP(sender) + "slapping players is not allowed");
                return;
            }
            else if (slap_list.Count == 0)
            {
                SendPlayerMessage(sender, PP(sender) + "server does not know any slaps");
                return;
            }


            String best_match = bestPlayerMatch(sender, victim, true, true);
            if (best_match == null || best_match.Length == 0)
            {
                SendConsoleWarn(sender, PP(sender) + "there is no player matching \"" + victim + "\"");
                return;
            }

            PlayerProfile pp = getPlayerProfile(best_match);

            if (pp == null)
            {
                SendConsoleWarn(sender, PP(sender) + "could not find profile for \"" + best_match + "\"");
                return;
            }

            Random rnd = new Random();


            string slap_text = slap_list[rnd.Next(slap_list.Count)];

            slap_text = slap_text.Replace("%sender%", ((sender == null) ? "Admin" : sender));
            slap_text = slap_text.Replace("%receiver%", pp.name);

            if (sender == null)
                ConsoleWrite(slap_text);

            SendGlobalMessage(slap_text);
        }

        private void insultPlayerCmd(string sender, string victim)
        {


            List<String> insult_list = getStringListVarValue("insult_msg_list");

            if (getBooleanVarValue("insult_player") == false)
            {
                SendPlayerMessage(sender, PP(sender) + "insulting players is not allowed");
                return;
            }
            else if (insult_list.Count == 0)
            {
                SendPlayerMessage(sender, PP(sender) + "server does not know any insults");
                return;
            }


            String best_match = bestPlayerMatch(sender, victim, true, true);
            if (best_match == null || best_match.Length == 0)
            {
                SendConsoleWarn(sender, PP(sender) + "there is no player matching \"" + victim + "\"");
                return;
            }

            PlayerProfile pp = getPlayerProfile(best_match);

            if (pp == null)
            {
                SendConsoleWarn(sender, PP(sender) + "could not find profile for \"" + best_match + "\"");
                return;
            }


            Random rnd = new Random();

            string insult_text = insult_list[rnd.Next(insult_list.Count)];

            insult_text = insult_text.Replace("%name%", pp.name);

            int SquadId = pp.info.SquadID;
            int TeamId = pp.info.TeamID;

            if (sender != null)
            {

                PlayerProfile sp = getPlayerProfile(sender);
                if (sp == null)
                {
                    SendConsoleWarn(sender, PP(sender) + "could not find your player profile");
                    return;
                }

                SquadId = sp.info.SquadID;
                TeamId = sp.info.TeamID;
            }


            // if sender and victim are in same Team/Squad, send only one message
            if (!(TeamId == pp.info.TeamID && SquadId == pp.info.SquadID))
                SendPlayerMessage(sender, insult_text);

            if (sender == null)
                ConsoleWrite(insult_text);

            SendPlayerMessage(pp.name, ((sender == null) ? "" : sender + " says, ") + insult_text);
        }

        private void mockPlayerSuicide(string name)
        {
            if (!getBooleanVarValue("mock_suicide"))
                return;

            List<String> mock_list = getStringListVarValue("mock_msg_list");

            if (mock_list.Count == 0)
                return;

            Random rnd = new Random();

            string mock_text = mock_list[rnd.Next(mock_list.Count)];
            mock_text = mock_text.Replace("%name%", name);

            SendPlayerMessage(name, mock_text);
        }

        private void myStatsCmd(string sender)
        {
            playerStatsCmd(sender, sender);
        }

        private void enableVarGroupCmd(string sender, string group)
        {
            enablePluginVarGroup(sender, group);
        }

        private void disableVarGroupCmd(string sender, string group)
        {
            disablePluginVarGroup(sender, group);
        }

        private bool setPluginVarGroup(string sender, string group, string val)
        {
            if (group == null)
                return false;

            group = group.Replace(";", ",");

            List<string> vars = new List<string>(Regex.Split(group, @"\s*,\s*", RegexOptions.IgnoreCase));
            foreach (string var in vars)
            {
                if (setPluginVarValue(sender, var, val))
                    SendPlayerMessage(sender, PP(sender) + var + " set to \"" + val + "\"");

            }

            return true;
        }

        private bool enablePluginVarGroup(string sender, string group)
        {
            //search for all variables matching
            List<string> vars = getVariableNames(group);
            if (vars.Count == 0)
            {
                DebugWrite("unknown variable \"" + group + "\"", 3);
                return false;
            }

            return setPluginVarGroup(sender, String.Join(",", vars.ToArray()), "true");
        }

        private List<string> getVariableNames(string group)
        {
            List<string> names = new List<string>();
            List<string> list = new List<string>(Regex.Split(group, @"\s*,\s*"));
            List<string> vars = getPluginVars();
            foreach (string search in list)
            {
                foreach (string var in vars)
                {
                    if (var.Contains(search))
                        if (!names.Contains(var))
                            names.Add(var);
                }
            }

            return names;
        }

        private bool disablePluginVarGroup(string sender, string group)
        {
            //search for all variables matching
            List<string> vars = getVariableNames(group);
            if (vars.Count == 0)
            {
                DebugWrite("unknown variable \"" + group + "\"", 3);
                return false;
            }

            return setPluginVarGroup(sender, String.Join(",", vars.ToArray()), "false");
        }

        private void getVariableCmd(string sender, string var)
        {
            string val = getPluginVarValue(sender, var);
            if (val != null)
                SendPlayerMessage(sender, PP(sender) + var + " = " + val);
        }

        public ViolationType getViolationType(string str)
        {
            if (str == null)
                return ViolationType.invalid;

            if (Regex.Match(str, @"\s*team_kill\s*", RegexOptions.IgnoreCase).Success)
                return ViolationType.team_kill;
            else
                return ViolationType.invalid;
        }

        public ActionType getActionType(string str)
        {
            if (str == null)
                return ActionType.invalid;

            if (Regex.Match(str, @"\s*votekick\s*", RegexOptions.IgnoreCase).Success)
                return ActionType.kick;
            else if (Regex.Match(str, @"\s*voteban\s*", RegexOptions.IgnoreCase).Success)
                return ActionType.ban;
            else if (Regex.Match(str, @"\s*punish\s*", RegexOptions.IgnoreCase).Success)
                return ActionType.punish;
            else if (Regex.Match(str, @"\s*forgive\s*", RegexOptions.IgnoreCase).Success)
                return ActionType.forgive;
            else
                return ActionType.invalid;
        }

        private void setVariableCmd(string sender, string var, string val)
        {
            if (setPluginVarValue(sender, var, val))
                SendPlayerMessage(sender, PP(sender) + var + " set to \"" + val + "\"");
        }

        private void pluginSettingsCmd(string sender)
        {
            SendConsoleMessage(sender, " == Insane Punisher Settings == ");
            foreach (string var in getPluginVars(false))
            {
                /* do not show advanced variables in the in-game list */
                if (this.hiddenVariables.Contains(var))
                    continue;

                SendConsoleMessage(sender, var + " = " + getPluginVarValue(sender, var));
            }
        }








        public bool setPluginVarValue(string var, string val)
        {
            return setPluginVarValue(null, var, val, false);
        }

        public bool setPluginVarValue(string var, string val, bool ui)
        {
            return setPluginVarValue(null, var, val, ui);
        }

        public bool setPluginVarValue(string sender, string var, string val)
        {
            return setPluginVarValue(sender, var, val, false);
        }



        private bool setPluginVarValue(string sender, string var, string val, bool ui)
        {
            if (var == null || val == null || !getPluginVars().Contains(var))
                return false;

            if (!getPluginVars().Contains(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return false;
            }

            /* Parse Boolean Values */
            bool booleanValue = false;
            bool isBooleanValue = true;
            if (Regex.Match(val, @"\s*(1|true|yes)\s*", RegexOptions.IgnoreCase).Success)
                booleanValue = true;
            else if (Regex.Match(val, @"\s*(0|false|no)\s*", RegexOptions.IgnoreCase).Success)
                booleanValue = false;
            else
                isBooleanValue = false;


            /* Parse Integer Values */
            int integerValue = 0;
            bool isIntegerValue = int.TryParse(val, out integerValue) && integerValue >= 0;


            /* Parse Float Values */
            float floatValue = 0F;
            bool isFloatValue = float.TryParse(val, out floatValue) && floatValue >= 0F;

            /* Parse String List */
            List<string> stringListValue = new List<string>(Regex.Split(val, @"\s*\|\s*"));
            stringListValue.RemoveAll(delegate(String item) { return item == null || item.Trim().Length == 0; });
            bool isStringList = true;

            /* Parse String var */
            string stringValue = val;
            bool isStringValue = (val != null);

            /* Parse Enum var */
            string enumValue = val;
            bool isEnumValue = (val != null);


            if (isBooleanVar(var))
            {
                if (!isBooleanValue)
                {
                    SendConsoleWarn(sender, PP(sender) + "\"" + val + "\" is invalid for " + var);
                    return false;
                }

                setBooleanVarValue(var, booleanValue);
                return true;
            }
            else if (isIntegerVar(var))
            {
                if (!isIntegerValue)
                {
                    SendConsoleWarn(sender, PP(sender) + "\"" + val + "\" is invalid for " + var);
                    return false;
                }

                setIntegerVarValue(var, integerValue);
                return true;
            }
            else if (isFloatVar(var))
            {
                if (!isFloatValue)
                {
                    SendConsoleWarn(sender, PP(sender) + "\"" + val + "\" is invalid for " + var);
                    return false;
                }

                setFloatVarValue(var, floatValue);
                return true;
            }
            else if (isStringListVar(var))
            {
                if (!isStringList)
                {
                    SendConsoleWarn(sender, PP(sender) + "\"" + val + "\"  is invalid for " + var);
                    return false;
                }

                setStringListVarValue(var, stringListValue);
                return true;
            }
            else if (isStringVar(var))
            {
                if (var.Equals("console"))
                {
                    PluginCommand(null, val);
                    return false;
                }

                if (!isStringValue)
                {
                    SendConsoleWarn(sender, PP(sender) + "\"" + val + "\"  is invalid for " + var);
                    return false;
                }

                setStringVarValue(var, stringValue);
                return true;
            }
            else if (isEnumVar(var))
            {
                if (!isEnumValue)
                {
                    SendConsoleWarn(sender, PP(sender) + "\"" + val + "\"  is invalid for " + var);
                    return false;
                }

                setEnumVarValue(var, enumValue);
                return true;
            }
            else
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return false;
            }

        }

        private bool isIntegerVar(string var)
        {
            return this.integerVariables.ContainsKey(var);
        }

        private int getIntegerVarValue(string var)
        {
            if (!isIntegerVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return -1;
            }

            return this.integerVariables[var];
        }

        private bool setIntegerVarValue(string var, int val)
        {
            if (!isIntegerVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return false;
            }

            this.integerVariables[var] = val;
            return true;
        }


        private bool isStringVar(string var)
        {
            return this.stringVariables.ContainsKey(var);
        }


        private string getStringVarValue(string var)
        {
            if (!isStringVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return "";
            }

            return this.stringVariables[var];
        }

        private bool setStringVarValue(string var, string val)
        {
            if (!isStringVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return false;
            }

            string oldval = this.stringVariables[var];
            this.stringVariables[var] = val;

            return true;
        }

        private bool isStringListVar(string var)
        {
            return this.stringListVariables.ContainsKey(var);
        }

     

        private List<string> getStringListVarValue(string var)
        {
            return getStringListVarValue(var, true);
        }

        private List<string> getStringListVarValue(string var, bool decode)
        {
            if (!isStringListVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return new List<string>();
            }

            string[] out_list = this.stringListVariables[var].Split('|');

            if (decode)
                for (int i = 0; i < out_list.Length; i++)
                    out_list[i] = CPluginVariable.Decode(out_list[i]);


            return new List<string>(out_list);
        }

        private bool setStringListVarValue(string var, List<string> val)
        {
            if (!isStringListVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return false;
            }

            List<string> cleanList = new List<string>();
            foreach (string item in val)
                if (Regex.Match(item, @"^\s*$").Success)
                    continue;
                else
                    cleanList.Add(item);

            //this.stringListVariables[var] = val;
            this.stringListVariables[var] = String.Join("|", val.ToArray());
            return true;
        }


        private bool isFloatVar(string var)
        {
            return this.floatVariables.ContainsKey(var);
        }

        private float getFloatVarValue(string var)
        {
            if (!isFloatVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return -1F;
            }

            return this.floatVariables[var];
        }

        private bool setFloatVarValue(string var, float val)
        {
            if (!isFloatVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return false;
            }

            this.floatVariables[var] = val;
            return true;
        }


        private bool isBooleanVar(string var)
        {
            return this.booleanVariables.ContainsKey(var);
        }

        private bool getBooleanVarValue(string var)
        {
            if (!isBooleanVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return false;
            }

            return this.booleanVariables[var];
        }

        private bool setBooleanVarValue(string var, bool val)
        {
            if (!isBooleanVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return false;
            }

            this.booleanVariables[var] = val;
            return true;
        }


        private bool isEnumVar(string var)
        {
            return this.enumVariables.ContainsKey(var);
        }


        private string getEnumVarValue(string var)
        {
            if (!isEnumVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return "";
            }

            return this.enumVariables[var];
        }

        private Object getEnumVarValueObj(string var)
        {
            if (!isEnumVar(var))
                return null;

            /* get the enum type */
            Type type = getEnumVarType(var);
            if (type == null)
                return null;

            /* get the enum value */
            string value = getEnumVarValue(var);
            if (value.Length == 0)
                return null;

            return System.Enum.Parse(type, value, true);
        }

        private bool setEnumVarValue(string var, string val)
        {
            if (!isEnumVar(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return false;
            }


            string oldval = this.enumVariables[var];
            this.enumVariables[var] = val;

            return true;
        }

        public Type getEnumVarType(string var)
        {
            if (!isEnumVar(var) || !this.enumVariablesType.ContainsKey(var))
                return null;

            return this.enumVariablesType[var];
        }



        private string getPluginVarValue(string var)
        {
            return getPluginVarValue(null, var, false);
        }

        private string getPluginVarValue(string var, bool ui)
        {
            return getPluginVarValue(null, var, ui);
        }

        private string getPluginVarValue(string sender, string var)
        {
            return getPluginVarValue(sender, var, false);
        }

        private string getPluginVarValue(string sender, string var, bool ui)
        {
            if (!getPluginVars().Contains(var))
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return "";
            }

            if (isBooleanVar(var))
            {
                return getBooleanVarValue(var).ToString();
            }
            else if (isIntegerVar(var))
            {
                return getIntegerVarValue(var).ToString();
            }
            else if (isFloatVar(var))
            {
                return getFloatVarValue(var).ToString();
            }
            else if (isStringListVar(var))
            {
                string lst = String.Join("|", getStringListVarValue(var, false).ToArray());
                return lst;
            }
            else if (isStringVar(var))
            {
                return getStringVarValue(var);
            }
            else if (isEnumVar(var))
            {
                return getEnumVarValue(var);
            }
            else
            {
                DebugWrite("unknown variable \"" + var + "\"", 3);
                return "";
            }
        }


        private string getPluginVarGroup(string varname)
        {
            if (varname == null || varname.Length == 0)
                return "Settings";

            foreach (KeyValuePair<string, List<string>> pair in variableGroups)
                foreach (string cmp_varname in pair.Value)
                    if (cmp_varname.CompareTo(varname) == 0)
                        return pair.Key;

            return "Settings";
        }


        private List<string> getPluginVars()
        {
            return getPluginVars(true);
        }






        private List<string> getPluginVars(bool show_advanced)
        {
            List<string> vars = new List<string>();
            vars.AddRange(getEnumPluginVars());
            vars.AddRange(getIntegerPluginVars());
            vars.AddRange(getBooleanPluginVars());
            vars.AddRange(getStringListPluginVars());
            vars.AddRange(getFloatPluginVars());
            vars.AddRange(getStringPluginVars());


            if (!(show_advanced))
            {
                foreach (string advanced_var in variableGroups["Advanced"])
                    vars.Remove(advanced_var);
            }


            return vars;
        }

        private List<string> getEnumPluginVars()
        {
            return new List<string>(this.enumVariables.Keys);
        }

        private List<string> getStringPluginVars()
        {
            return new List<string>(this.stringVariables.Keys);
        }


        private List<string> getStringListPluginVars()
        {
            return new List<string>(this.stringListVariables.Keys);
        }


        private List<string> getIntegerPluginVars()
        {
            return new List<string>(this.integerVariables.Keys);
        }

        private List<string> getFloatPluginVars()
        {
            return new List<string>(this.floatVariables.Keys);
        }

        private List<string> getBooleanPluginVars()
        {
            return new List<string>(this.booleanVariables.Keys);
        }



        private List<PlayerProfile> getVictimAndKillerProfiles(string victimName, string killerName, ActionType action)
        {
            if (victimName == null)
                return null;



            List<PlayerProfile> profiles = new List<PlayerProfile>();

            //get the victim profile
            PlayerProfile victimProfile = getPlayerProfile(victimName);
            if (victimProfile == null)
            {
                SendPlayerMessage(victimName, PP(victimName) + "server cannot find your player profile");
                return null;
            }

            //get the killer's profile
            PlayerProfile killerProfile = null;

            if (killerName == null)
            {
                if (victimProfile.getLastKiller() == null)
                    SendPlayerMessage(victimName, PP(victimName) + "there is no-one to " + action.ToString() + " at this time");
                else
                    killerProfile = victimProfile.getLastKiller();
            }
            else
            {
                String best_match = bestPlayerMatch(victimName, killerName, true, true);
                if (best_match == null || best_match.Length == 0)
                    SendConsoleWarn(victimName, PP(victimName) + "there is no player matching \"" + killerName + "\"");
                else
                    killerProfile = getPlayerProfile(best_match);
            }

            if (killerProfile == null)
            {
                SendConsoleWarn(victimName, PP(victimName) + "could not find profile for \"" + killerName + "\"");
                return null;
            }

            profiles.Add(victimProfile);
            profiles.Add(killerProfile);

            return profiles;
        }


        private void punishMyKillerCmd(string victimName, string killerName)
        {
            punishMyKillerCmd(victimName, killerName, false);
        }

        private void punishMyKillerCmd(string victimName)
        {
            punishMyKillerCmd(victimName, null, false);
        }

        private void punishMyKillerCmd(string victimName, Boolean auto)
        {
            punishMyKillerCmd(victimName, null, auto);
        }

        private void punishMyKillerCmd(string victimName, string killerName, Boolean auto)
        {
            if (getActionableViolations(ActionType.punish).Count == 0)
            {
                SendPlayerMessage(victimName, PP(victimName) + action2stringING(ActionType.punish) + " is not allowed");
                return;
            }

            List<PlayerProfile> profiles = getVictimAndKillerProfiles(victimName, killerName, ActionType.punish);

            if (profiles == null)
                return;

            PlayerProfile victimProfile = profiles[0];
            profiles.RemoveAt(0);

            List<PlayerProfile> killerProfiles = profiles;


            //there is one or more killers under that name

            foreach (PlayerProfile killerProfile in killerProfiles)
                if (victimProfile.punish(killerProfile, auto))
                    enforcePendingAutoKicks(killerProfile);
        }


        private void voteMyKillerCmd(string victimName)
        {
            voteMyKillerCmd(victimName, null);
        }

        private void voteMyKillerCmd(string victimName, string killerName)
        {
           
            if (getActionableViolations(ActionType.kick).Count == 0)
            {
                //SendPlayerMessage(victimName, PP(victimName) + action2stringING(ActionType.kick) + " is not allowed");
                return;
            }
             

            List<PlayerProfile> profiles = getVictimAndKillerProfiles(victimName, killerName, ActionType.kick);

            if (profiles == null)
                return;

            PlayerProfile victimProfile = profiles[0];
            profiles.RemoveAt(0);

            List<PlayerProfile> killerProfiles = profiles;

            //there is one or more killers under that name
            foreach (PlayerProfile killerProfile in killerProfiles)
                if (victimProfile.vote(killerProfile) > 0)
                    enforcePendingKicks(killerProfile);
        }


        private void enforcePendingKicks(PlayerProfile killer)
        {
            int votesRec = killer.getVotesReceived();
            string violations = killer.getViolationsStr();


            if (getBooleanVarValue("kick_team_kill"))
            {
                /* vote kicking is enabled for at least one type of violation */
                if (votesRec > getIntegerVarValue("max_kick_votes"))
                {
                    SendGlobalMessage(killer.name + " was vote kicked for " + violations);
                    KickPlayerWithMessage(killer.name, " vote kicked for " + violations);
                }
                else
                {
                    if (votesRec == getIntegerVarValue("max_kick_votes"))
                        SendGlobalMessage(killer.name + " is one vote away from being vote kicked!");
                    else
                        SendPlayerMessage(killer.name, PP(killer.name) + "you are " + (getIntegerVarValue("max_kick_votes") - votesRec + 1) + " votes away from being vote kicked!");
                }
            }
        }


        private void enforcePendingAutoKicks(PlayerProfile killer)
        {
            string violations = killer.getViolationsStr();
            int auto_max = getIntegerVarValue("auto_action_max");
            AutoAction auto_action = (AutoAction)getEnumVarValueObj("auto_action");
            int vcount = (killer.getViolations() - killer.getPendingViolations()); /* do not count pending violations */

            if (!auto_action.Equals(AutoAction.None) && auto_max > 0)
            {
                String auto_action_str = autoaction2stringED(auto_action);

                /* auto kicking is enabled */
                if (vcount > auto_max)
                {
                    SendGlobalMessage(killer.name + " was auto " + auto_action_str + " for " + violations);
                    if (auto_action.Equals(AutoAction.Kick))
                        KickPlayerWithMessage(killer.name, " auto " + auto_action_str + " for " + violations);
                    else if (auto_action.Equals(AutoAction.Ban))
                        BanPlayerWithMessage((BanType)getEnumVarValueObj("ban_type"), (BanDuration)getEnumVarValueObj("ban_duration"), killer.name, getIntegerVarValue("ban_minutes"), " vote banned for " + violations);
                }
                else if (vcount == auto_max)
                    SendGlobalMessage(killer.name + " is one violation away from being auto " + auto_action_str + "!");
                else
                    SendPlayerMessage(killer.name, PP(killer.name) + "you are " + (auto_max - vcount) + " violations away from being auto " + auto_action_str + "!");
            }
        }

        private void enforcePendingBans(PlayerProfile killer)
        {
            int votesRec = killer.getBanVotesReceived();
            string violations = killer.getViolationsStr();


            if (votesRec > getIntegerVarValue("max_ban_votes"))
            {
                SendGlobalMessage(killer.name + " was vote banned for " + violations);
                int time = getIntegerVarValue("ban_minutes");
                BanPlayerWithMessage((BanType)getEnumVarValueObj("ban_type"), (BanDuration)getEnumVarValueObj("ban_duration"), killer.name, time, " vote banned for " + violations);
            }
            else
            {
                if (votesRec == getIntegerVarValue("max_ban_votes"))
                    SendGlobalMessage(killer.name + " is one vote away from being banned!");
                else
                    SendPlayerMessage(killer.name, killer.name + "you are " + (getIntegerVarValue("max_ban_votes") - votesRec + 1) + " votes away from being banned!");
            }
        }


        private void banMyKillerCmd(string victimName)
        {
            banMyKillerCmd(victimName, null);
        }

        private void banMyKillerCmd(string victimName, string killerName)
        {
            
            if (getActionableViolations(ActionType.ban).Count == 0)
            {
                //SendPlayerMessage(victimName, PP(victimName) + action2stringING(ActionType.ban) + " is not allowed");
                return;
            }

            List<PlayerProfile> profiles = getVictimAndKillerProfiles(victimName, killerName, ActionType.ban);

            if (profiles == null)
                return;

            PlayerProfile victimProfile = profiles[0];
            profiles.RemoveAt(0);

            List<PlayerProfile> killerProfiles = profiles;

            //there is one or more killers under that name
            foreach (PlayerProfile killerProfile in killerProfiles)
                if (victimProfile.ban(killerProfile) > 0)
                    enforcePendingBans(killerProfile);
        }

        private void forgiveMyKillerCmd(string victimName)
        {
            forgiveMyKillerCmd(victimName, null);

        }

        private void forgiveMyKillerCmd(string victimName, string killerName)
        {

            if (getActionableViolations(ActionType.forgive).Count == 0)
            {
                SendPlayerMessage(victimName, action2stringING(ActionType.forgive) + " is not allowed");
                return;
            }

            List<PlayerProfile> profiles = getVictimAndKillerProfiles(victimName, killerName, ActionType.forgive);

            if (profiles == null)
                return;

            PlayerProfile victimProfile = profiles[0];
            profiles.RemoveAt(0);

            List<PlayerProfile> killerProfiles = profiles;

            //there is one or more killers under that name
            foreach (PlayerProfile killerProfile in killerProfiles)
                victimProfile.forgive(killerProfile);

        }


        public string violation2stringING(ViolationType violation)
        {
            switch (violation)
            {
                case ViolationType.team_kill:
                    return "team killing";
                default:
                    return "(%violation%)";
            }
        }
        public string playerstate2stringED(PlayerState state)
        {
            switch (state)
            {
                case PlayerState.alive:
                    return "is alive";
                case PlayerState.dead:
                    return "is dead";
                case PlayerState.kicked:
                    return "was kicked";
                case PlayerState.banned:
                    return "was banned";
                case PlayerState.left:
                    return "left the game";
                case PlayerState.limbo:
                    return "is in limbo";
                case PlayerState.violated:
                    return "was violated";
                default:
                    ConsoleWarn("unknown value " + state.ToString() + " type for " + typeof(PlayerState).Name);
                    return "(%player_state%)";
            }

        }

        public string autoaction2stringED(AutoAction action)
        {
            switch (action)
            {
                case AutoAction.Ban:
                    return "banned";
                case AutoAction.Kick:
                    return "kicked";
                default:
                    ConsoleWarn("unknown value " + action.ToString() + " type for " + typeof(AutoAction).Name);
                    return "(%auto_action%)";
            }

        }


        public string action2stringING(ActionType action)
        {
            switch (action)
            {
                case ActionType.ban:
                    return "banning";
                case ActionType.forgive:
                    return "forgiving";
                case ActionType.punish:
                    return "punishing";
                case ActionType.kick:
                    return "kicking";
                default:
                    ConsoleWarn("unknown value " + action.ToString() + " type for " + typeof(ActionType).Name);
                    return "(%action%)";
            }
        }

        public string list2string(List<string> list, string glue)
        {

            if (list == null || list.Count == 0)
                return "";
            else if (list.Count == 1)
                return list[0];

            string last = list[list.Count - 1];
            list.RemoveAt(list.Count - 1);

            string str = "";
            foreach (string item in list)
                str += item + ", ";

            return str + glue + last;
        }

        public string list2string(List<string> list)
        {
            return list2string(list, "and ");
        }


        private void playersStatsCmd(string sender)
        {
            SendPlayerMessage(sender, "== Insane Punisher Violations ==");
            int count = 0;
            
            bool long_format = false;

            if (sender == null)
                long_format = true;

            foreach (KeyValuePair<string, PlayerProfile> pair in this.players)
            {
                count++;
                PlayerProfile playerProfile = pair.Value;
                SendPlayerMessage(sender, count + ". " + playerProfile.ToString(long_format));
            }
        }

        // setup personalized prefix
        public string PP(String sender)
        {
            if (sender == null)
                return "";
            else
                return sender + ", ";
        }

        private void playerStatsCmd(string sender, string name)
        {

            if (name == null)
                return;

            String best_match = bestPlayerMatch(sender, name, true, true);
            if (best_match == null || best_match.Length == 0)
            {
                SendConsoleWarn(sender, PP(sender) + "there is no player matching \"" + name + "\"");
                return;
            }

            PlayerProfile killer = getPlayerProfile(best_match);

            if (killer == null)
            {
                SendConsoleWarn(sender, PP(sender) + "could not find profile for \"" + best_match + "\"");
                return;
            }

            List<PlayerProfile> voters = killer.getVoters();
            List<PlayerProfile> banvoters = killer.getBanVoters();
            List<PlayerProfile> punishers = killer.getQueuedPunishersList();
            List<PlayerProfile> teamKillVictims = killer.getTeamKillVictims();


            SendPlayerMessage(sender, "=== " + killer.name + " ( " + killer.state.ToString() + " ) ===");
            SendPlayerMessage(sender, "    Queued Punishments: " + killer.getQueuedPunishmentsCount());
            if (sender == null)
                foreach (PlayerProfile victim in punishers)
                    SendPlayerMessage(sender, "        " + victim.name + " (" + killer.getQueuedPunishmentsStr(victim) + ")");

            SendPlayerMessage(sender, "   Kick Votes Received: " + killer.getVotesReceived());
            if (sender == null)
                foreach (PlayerProfile victim in voters)
                    SendPlayerMessage(sender, "        " + victim.name + " (" + killer.getVotesReceived(victim) + ")");


            SendPlayerMessage(sender, "   Ban Votes Received: " + killer.getBanVotesReceived());
            if (sender == null)
                foreach (PlayerProfile victim in banvoters)
                    SendPlayerMessage(sender, "        " + victim.name + " (" + killer.getBanVotesReceived(victim) + ")");
 

            SendPlayerMessage(sender, "    Team Kills: " + killer.getTeamKills());
            if (sender == null)
                foreach (PlayerProfile victim in teamKillVictims)
                    SendPlayerMessage(sender, "        " + victim.name + " (Total: " + killer.getTeamKills(victim) + ", Pending: " + killer.getPendingTeamKills(victim) + ") ");
        }


        private bool isAdmin(string soldier)
        {
            if (soldier == null || soldier.Trim().Length == 0)
                return false;

            /* For Debug, I am always Admin */
            if (soldier.Equals("micovery"))
                return true;



            List<string> admin_list = getStringListVarValue("admin_list");
            return admin_list.Contains(soldier);
        }

        private bool isImmune(PlayerProfile player)
        {
            if (player == null)
                return false;

            string name = player.name;

            List<string> player_white_list = getStringListVarValue("player_white_list");

            List<string> immune = new List<string>();
            immune.AddRange(player_white_list);

            if (immune.Contains(name))
                return true;

            return false;
        }



        private void TeamKill(CPlayerInfo killer, CPlayerInfo victim)
        {

            PlayerProfile killerProfile = getPlayerProfile(killer);
            PlayerProfile victimProfile = getPlayerProfile(victim);

            if (killerProfile != null && victimProfile != null)
                killerProfile.addTk(victimProfile);

            if (getBooleanVarValue("auto_punish_" + ViolationType.team_kill.ToString()))
                punishMyKillerCmd(victim.SoldierName, true);

        }


        private List<string> getAllowedActionCommands(ViolationType violation)
        {
            List<string> list = new List<string>();

            if (violation.Equals(ViolationType.invalid))
                return list;

            foreach (ActionType action in Enum.GetValues(typeof(ActionType)))
            {
                if (action.Equals(ActionType.invalid))
                    continue;

                string var = getActionViolationVarName(action, violation);
                bool val = getBooleanVarValue(var);

                if (val)
                    list.Add(getActionCommand(action));
            }

            return list;
        }


        private string getViolationInfoMessage(PlayerProfile killer, PlayerProfile victim, ViolationType violation)
        {
            string info_msg = "";
            if (violation.Equals(ViolationType.team_kill))
                info_msg = PP(victim.name) + killer.name + " team killed you!";

            return info_msg;

        }

        private void showActionCommands(PlayerProfile victim, PlayerProfile killer, ViolationType violation)
        {

            if (violation.Equals(ViolationType.invalid))
                return;

            string warn_msg = getWarningMessage(killer, victim, violation);
            string info_msg = getViolationInfoMessage(killer, victim, violation);
            string act_msg = getActionCommandsStr(violation, victim, killer);

            //if there is no info message, or action message quit
            if (act_msg.CompareTo("") == 0 || info_msg.CompareTo("") == 0)
                return;


            bool say = getBooleanVarValue("say_" + violation.ToString());
            bool auto_punish = getBooleanVarValue("auto_punish_" + violation.ToString());


            //there are no actions allowed for this voilation


            if (say || auto_punish)
            {
                if (!auto_punish)
                {
                    SendPlayerMessage(victim.name, info_msg);
                    SendPlayerMessage(victim.name, act_msg);
                }

                //only send if there are any warnings
                if (warn_msg.Length > 0)
                    SendPlayerMessage(killer.name, warn_msg);
            }
        }

        private string getActionCommandsStr(ViolationType violation, PlayerProfile victim, PlayerProfile killer)
        {
            List<string> cmds = getAllowedActionCommands(violation);
            string msg = list2string(cmds, "or ");
            if (msg.CompareTo("") == 0)
                return "";
            else
                return PP(victim.name) + "Take action by typing: " + msg;
        }


        public string getActionViolationVarName(ActionType action, ViolationType violation)
        {
            return action.ToString() + "_" + violation.ToString();
        }

        public List<ViolationType> getActionableViolations(ActionType action)
        {
            List<ViolationType> list = new List<ViolationType>();


            if (action.Equals(ActionType.invalid))
                return list;

            foreach (ViolationType violation in Enum.GetValues(typeof(ViolationType)))
            {
                if (violation.Equals(ViolationType.invalid))
                    continue;

                string var = getActionViolationVarName(action, violation);

                if (getBooleanVarValue(var) || (action.Equals(ActionType.punish) && getBooleanVarValue("auto_" + var)))
                    list.Add(violation);
            }

            return list;
        }

        public string getActionableViolationsStr(ActionType action)
        {
            List<ViolationType> violations = getActionableViolations(action);

            List<String> list = new List<string>();
            foreach (ViolationType violation in violations)
                list.Add(violation2stringING(violation));

            return list2string(list, "or ");
        }

        private string getActionCommand(ActionType action)
        {
            if (action.Equals(ActionType.forgive))
                return "!forgive";
            else if (action.Equals(ActionType.punish))
                return "!punish";
            else if (action.Equals(ActionType.kick))
                return "!votekick";
            else if (action.Equals(ActionType.ban))
                return "!voteban";

            return "!votekick";
        }



        private List<string> getViolationWarnings(ViolationType violation)
        {
            if (violation.Equals(ViolationType.team_kill))
                return getStringListVarValue("team_kill_warn_list");
            else
                return new List<string>();
        }

        private string getWarningMessage(PlayerProfile killer, PlayerProfile victim, ViolationType violation)
        {


            List<string> list = getViolationWarnings(violation);
            int count = killer.getViolations(violation);
            int total = list.Count;

            string msg;

            if (list.Count == 0)
                msg = "";
            else if (count >= total)
                msg = list[total - 1];
            else
                msg = list[count - 1];

            //do the replacements

            msg = msg.Replace("%killer%", killer.name);
            msg = msg.Replace("%victim%", victim.name);
            msg = msg.Replace("%count%", count.ToString());

            return msg;
        }


        private PlayerProfile getPlayerProfile(CPlayerInfo info)
        {
            return getPlayerProfile(info.SoldierName);
        }


        private PlayerProfile getPlayerProfile(string name)
        {
            if (name == null)
                return null;

            PlayerProfile pp;
            this.players.TryGetValue(name, out pp);

            // try updating the player's 
            if (pp == null)
            {
                getPlayersList();
                getPBPlayersList();
            }
            return pp;
        }



        public void getPlayersList()
        {
            ServerCommand("admin.listPlayers", "all");

        }

        public void getPBPlayersList()
        {
            ServerCommand("punkBuster.pb_sv_command", "pb_sv_plist");
        }



        Object mutex = new Object();
        public void syncPlayersList(List<CPlayerInfo> lstPlayers)
        {

            lock (mutex)
            {
                // first update the information that players that still are in list
                foreach (CPlayerInfo cpiPlayer in lstPlayers)
                    if (this.players.ContainsKey(cpiPlayer.SoldierName))
                        this.players[cpiPlayer.SoldierName].info = cpiPlayer;

                //build a lookup table
                Dictionary<String, bool> player_lookup = new Dictionary<string, bool>();
                foreach (CPlayerInfo pinfo in lstPlayers)
                    if (!player_lookup.ContainsKey(pinfo.SoldierName))
                        player_lookup.Add(pinfo.SoldierName, true);


                List<String> players_to_remove = new List<string>();

                // now make a list of players that will need to be removed
                foreach (KeyValuePair<String, PlayerProfile> pair in players)
                    if (!player_lookup.ContainsKey(pair.Key) && !players_to_remove.Contains(pair.Key))
                        players_to_remove.Add(pair.Key);



                // now actually remove them
                foreach (String pname in players_to_remove)
                    if (players.ContainsKey(pname))
                    {
                        PlayerProfile pp = getPlayerProfile(pname);
                        if (pp != null)
                            pp.state = PlayerState.left;
                        players.Remove(pname);
                    }
            }
        }

        public override void OnListPlayers(List<CPlayerInfo> lstPlayers, CPlayerSubset cpsSubset)
        {

            if (cpsSubset.Subset != CPlayerSubset.PlayerSubsetType.All)
                return;

            syncPlayersList(lstPlayers);

        }


        public void DumpException(Exception e)
        {
            DumpException(e, String.Empty);
        }

        public void DumpException(Exception e, String prefix)
        {
            try
            {
                string class_name = this.GetType().Name;
                string path = makeRelativePath(class_name + ".dump");


                if (prefix == null)
                    prefix = String.Empty;
                else
                    prefix += ": ";


                if (e.GetType().Equals(typeof(ThreadAbortException)))
                {
                    Thread.ResetAbort();
                    return;
                }
                else if (e.GetType().Equals(typeof(TargetInvocationException)) && e.InnerException != null)
                {
                    ConsoleException(prefix + e.InnerException.GetType() + ": " + e.InnerException.Message);
                    ConsoleWrite("^1Extra information dumped in file " + path);
                    DumpExceptionFile(e, path);
                    DumpExceptionFile(e.InnerException, path);
                }
                else
                {
                    ConsoleException(prefix + e.GetType() + ": " + e.Message);

                    foreach (DictionaryEntry de in e.Data)
                    {
                        ConsoleWrite("    " + de.Key.ToString() + ": " + de.Value.ToString());
                    }

                    ConsoleWrite("^1Extra information dumped in file " + path);
                    DumpExceptionFile(e, path);
                }
            }
            catch (Exception ex)
            {
                ConsoleWarn("unable to dump extra exception information.");
                ConsoleException(ex.GetType() + ": " + ex.Message);
            }
        }

        public void DumpExceptionFile(Exception e, string path)
        {
            string class_name = this.GetType().Name;

            using (FileStream fs = File.Open(path, FileMode.Append))
            {
                String version = GetPluginVersion();
                String trace_str = "\n-----------------------------------------------\n";
                trace_str += "Version: " + class_name + " " + version + "\n";
                trace_str += "Date: " + DateTime.Now.ToString() + "\n";
                trace_str += e.GetType() + ": " + e.Message + "\n\n";
                trace_str += "Stack Trace: \n" + e.StackTrace + "\n\n";
                trace_str += "MSIL Stack Trace:\n";

                StackTrace trace = new StackTrace(e);
                StackFrame[] frames = trace.GetFrames();
                foreach (StackFrame frame in frames)
                    trace_str += "    " + frame.GetMethod() + ", IL: " + String.Format("0x{0:X}", frame.GetILOffset()) + "\n";


                Byte[] info = new UTF8Encoding(true).GetBytes(trace_str);
                fs.Write(info, 0, info.Length);
            }

        }


        public static String makeRelativePath(String file)
        {
            String exe_path = Directory.GetParent(Application.ExecutablePath).FullName;
            String dll_path = Directory.GetParent(Assembly.GetExecutingAssembly().Location).FullName;

            String rel_path = dll_path.Replace(exe_path, "");
            rel_path = Path.Combine(rel_path.Trim(new char[] { Path.DirectorySeparatorChar, Path.AltDirectorySeparatorChar }), file);
            return rel_path;
        }

        public string bestPlayerMatch(String sender, String player, bool fuzzy, bool quiet)
        {

            try
            {
                if (player == null)
                    return null;
                if (players.ContainsKey(player))
                    return player;

                int edit_distance = 0;
                String new_player = null;
                if ((new_player = bestMatch(player, new List<string>(players.Keys), out edit_distance, fuzzy)) == null)
                {
                    if (!quiet)
                        SendConsoleWarn(sender, PP(sender) + "could not find ^b" + player + "^n");
                    return null;
                }
                if (!quiet)
                    SendConsoleWarn(sender, PP(sender) + "could not find ^b" + player + "^n, but found ^b" + new_player + "^n, with edit distance of ^b" + edit_distance + "^n");
                return new_player;
            }
            catch (Exception e)
            {
                DumpException(e);
            }

            return null;
        }

        //find the best match for name in names
        public string bestMatch(String name, List<String> names, out int best_distance, bool fuzzy)
        {

            best_distance = int.MaxValue;


            try
            {

                //do the obvious check first
                if (names.Contains(name))
                {
                    best_distance = 0;
                    return name;
                }

                //name is not in the list, find the best match
                String best_match = null;

                // first try to see if any of the names contains target name as substring, so we can reduce the search
                Dictionary<String, String> sub_names = new Dictionary<string, string>();

                String name_lower = name.ToLower();

                for (int i = 0; i < names.Count; i++)
                {
                    String cname = names[i].ToLower();
                    if (cname.Equals(name_lower))
                        return names[i];
                    else if (cname.Contains(name_lower) && !sub_names.ContainsKey(cname))
                        sub_names.Add(cname, names[i]);
                }

                if (sub_names.Count > 0)
                    names = new List<string>(sub_names.Keys);

                if (sub_names.Count == 1)
                {
                    // we can optimize, and exit early
                    best_match = sub_names[names[0]];
                    best_distance = Math.Abs(best_match.Length - name.Length);
                    return best_match;
                }

                // if we are not doing a fuzzy search, and we have not found more than one sub-string, we can exit here
                if (!fuzzy && sub_names.Count == 0)
                    return null;

                // find the best/fuzzy match using modified Leveshtein algorithm              
                foreach (String cname in names)
                {
                    int distance = LevenshteinDistance(name, cname);
                    if (distance < best_distance)
                    {
                        best_distance = distance;
                        best_match = cname;
                    }
                }

                if (best_match == null)
                    return null;

                best_distance += Math.Abs(name.Length - best_match.Length);

                // if we searched through sub-names, get the actual match
                if (sub_names.Count > 0 && sub_names.ContainsKey(best_match))
                    best_match = sub_names[best_match];

                return best_match;
            }
            catch (Exception e)
            {
                DumpException(e);
            }

            return null;
        }

        // modified algorithm to ignore insertions, and case
        public static int LevenshteinDistance(string s, string t)
        {
            s = s.ToLower();
            t = t.ToLower();

            int n = s.Length;
            int m = t.Length;
            int[,] d = new int[n + 1, m + 1];

            if (n == 0)
                return m;

            if (m == 0)
                return n;

            for (int i = 0; i <= n; d[i, 0] = i++) ;
            for (int j = 0; j <= m; d[0, j] = j++) ;



            for (int i = 1; i <= n; i++)
                for (int j = 1; j <= m; j++)
                    d[i, j] = Math.Min(Math.Min(d[i - 1, j] + 1, d[i, j - 1] + 0), d[i - 1, j - 1] + ((t[j - 1] == s[i - 1]) ? 0 : 1));

            return d[n, m];
        }


        public static string Encode(string str)
        {
            byte[] encbuff = System.Text.Encoding.UTF8.GetBytes(str);
            return Convert.ToBase64String(encbuff);
        }
        public static string Decode(string str)
        {
            byte[] decbuff = Convert.FromBase64String(str.Replace(" ", "+"));
            return System.Text.Encoding.UTF8.GetString(decbuff);
        }

    }

    
}